from typing import Any, Callable, Optional, Union, Tuple, Hashable, Sequence

try:
    from collections import Mapping
except ImportError:
    # For python > 3.10
    from collections.abc import Mapping

from gelidum.exceptions import FrozenException
from gelidum.frozen import FrozenBase
from gelidum.typing import FrozenType, FrozenDict

__all__ = ["frozendict"]


class frozendict(dict, FrozenBase):  # noqa
    def __raise_immutable_exception(self, *args, **kwargs):
        raise FrozenException("'frozendict' object is immutable")

    def __init__(
        self,
        seq: Optional[Union[Mapping, Sequence, Tuple[Hashable, Any]]] = None,
        freeze_func: Optional[Callable[[Any], FrozenBase]] = None,
        **kwargs,
    ):
        if freeze_func is None:

            def freeze_func(item: Any) -> FrozenType:
                from gelidum.freeze import freeze

                return freeze(item, on_update="exception", on_freeze="copy")

        if seq is not None:
            items = None
            if isinstance(seq, Mapping):
                items = seq.items()
            elif isinstance(seq, Sequence):
                items = seq

            if items is not None:
                super().__init__(
                    {key: freeze_func(value) for key, value in items},
                    **{key: freeze_func(value) for key, value in kwargs.items()},
                )
            else:
                super().__init__({key: freeze_func(value) for key, value in seq})
        elif kwargs:
            super().__init__({key: freeze_func(value) for key, value in kwargs.items()})
        else:
            super().__init__()

    @classmethod
    def _gelidum_on_update(cls, *args, **kwargs):
        raise FrozenException("'frozendict' object is immutable")

    @classmethod
    def get_gelidum_hot_class_name(cls) -> str:
        return "dict"

    @classmethod
    def get_gelidum_hot_class_module(cls) -> str:
        return "builtins.dict"

    def __hash__(self) -> int:
        return hash(tuple((k, v) for k, v in self.items()))

    def __getitem__(self, key) -> Any:
        if type(key) is slice:
            return frozendict(super().__getitem__(key))
        try:
            return super().__getitem__(key)
        except IndexError:
            raise IndexError("frozendict index out of range")

    def __add__(self, other: FrozenDict) -> FrozenDict:
        joined_dict = self | other
        return frozendict(joined_dict)

    def __or__(self, other: FrozenDict) -> FrozenDict:
        if hasattr(super, "__or__"):
            return super().__or__(other)
        # Python version < 3.9
        result_dict = dict()
        result_dict.update(self)
        result_dict.update(other)
        return frozendict(result_dict)

    def __sub__(self, other: FrozenDict) -> FrozenDict:
        return frozendict({k: v for k, v in self.items() if k not in other})

    def remove(self, x):
        self.__raise_immutable_exception()

    def pop(self, items):
        self.__raise_immutable_exception()

    def popitem(self, *args, **kwarg):
        self.__raise_immutable_exception()

    def __setitem__(self, key, val, *args, **kwargs):
        self.__raise_immutable_exception()

    def __delitem__(self, key, *args, **kwargs):
        self.__raise_immutable_exception()

    def clear(self):
        self.__raise_immutable_exception()

    def update(self, *args, **kwarg):
        self.__raise_immutable_exception()

    def copy(self) -> "frozendict":
        """
        frozendict objects are only shallow-copied.
        """
        return self



import io
import sys
import warnings
from typing import List, Set, Dict, Any, Optional, Union, Tuple

from gelidum.dependencies import NUMPY_INSTALLED
from gelidum.collections import frozendict, frozenlist, frozenzet
from gelidum.exceptions import FrozenException
from gelidum.frozen import make_frozen_class, FrozenBase
from gelidum.frozen.frozen_class_creator import make_unique_class
from gelidum.typing import OnFreezeFuncType, OnUpdateFuncType, T, FrozenType, FrozenList
from gelidum.utils import isbuiltin
from gelidum.on_freeze import on_freeze_func_creator, OnFreezeCopier

if NUMPY_INSTALLED:
    from gelidum.collections import frozenndarray

NpArrayType = Any


def freeze(
    obj: T,
    on_update: Union[str, OnUpdateFuncType] = "exception",
    on_freeze: Union[str, OnFreezeFuncType] = "copy",
    save_original_on_copy: bool = False,
    inplace: Optional[bool] = None,
) -> FrozenType:

    # inplace argument will be removed from freeze in the next major version (0.6.0)
    if isinstance(inplace, bool):
        warnings.warn(
            DeprecationWarning("Use of inplace is deprecated and will be removed in next major version (0.6.0)")
        )

        if hasattr(obj.__class__, "__slots__") and inplace:
            raise FrozenException("Objects of classes with __slots__ cannot be frozen inplace")

        on_freeze_func: OnFreezeFuncType = on_freeze_func_creator(on_freeze="inplace" if inplace else "copy")

    else:
        if hasattr(obj.__class__, "__slots__") and on_freeze == "inplace":
            raise FrozenException("Objects of classes with __slots__ cannot be frozen inplace")

        on_freeze_func: OnFreezeFuncType = on_freeze_func_creator(on_freeze=on_freeze)

    on_update_func: OnUpdateFuncType = __on_update_func(on_update=on_update)

    return __freeze(
        obj=obj, on_update=on_update_func, on_freeze=on_freeze_func, save_original_on_copy=save_original_on_copy
    )


def __freeze(
    obj: Any, on_update: OnUpdateFuncType, on_freeze: OnFreezeFuncType, save_original_on_copy: bool = False
) -> Any:

    if isbuiltin(obj):
        return obj

    if isinstance(obj, FrozenBase):
        return obj

    class_name = type(obj).__name__
    freeze_func_name = f"__freeze_{class_name}"
    this_module = sys.modules[__name__]
    if hasattr(this_module, freeze_func_name):
        freeze_func = getattr(this_module, freeze_func_name)
        return freeze_func(obj, on_update=on_update, on_freeze=on_freeze)

    if NUMPY_INSTALLED:
        import numpy as np

        if isinstance(obj, np.ndarray):
            return __freeze_ndarray(obj, on_update=on_update, on_freeze=on_freeze)

    if isinstance(obj, object):
        return __freeze_object(
            obj, on_update=on_update, on_freeze=on_freeze, save_original_on_copy=save_original_on_copy
        )

    # Actually, this code is unreachable
    raise ValueError(f"object of type {obj.__class__} not frozen")  # pragma: no cover


def __freeze_bytearray(obj: bytearray, *args, **kwargs) -> bytes:  # noqa
    return bytes(obj)


def __freeze_ndarray(obj: NpArrayType, on_update: OnUpdateFuncType, on_freeze: OnFreezeFuncType) -> FrozenList:
    def freeze_func(item: Any) -> FrozenType:
        return freeze(item, on_update=on_update, on_freeze=on_freeze)

    return frozenndarray(obj, freeze_func=freeze_func)


def __freeze_dict(obj: Dict, on_update: OnUpdateFuncType, on_freeze: OnFreezeFuncType) -> frozendict:
    def freeze_func(item: Any) -> FrozenType:
        return freeze(item, on_update=on_update, on_freeze=on_freeze)

    return frozendict(obj, freeze_func=freeze_func)


def __freeze_list(obj: List, on_update: OnUpdateFuncType, on_freeze: OnFreezeFuncType) -> FrozenList:
    def freeze_func(item: Any) -> FrozenType:
        return freeze(item, on_update=on_update, on_freeze=on_freeze)

    return frozenlist(obj, freeze_func=freeze_func)


def __freeze_tuple(obj: Tuple, on_update: OnUpdateFuncType, on_freeze: OnFreezeFuncType) -> Tuple:
    return tuple(freeze(item, on_update=on_update, on_freeze=on_freeze) for item in obj)


def __freeze_set(obj: Set, on_update: OnUpdateFuncType, on_freeze: OnFreezeFuncType) -> frozenzet:
    def freeze_func(item: Any) -> FrozenType:
        return freeze(item, on_update=on_update, on_freeze=on_freeze)

    return frozenzet(obj, freeze_func=freeze_func)


def __freeze_TextIOWrapper(*args, **kwargs) -> None:  # noqa
    raise io.UnsupportedOperation("Text file handlers can't be frozen")


def __freeze_BufferedWriter(*args, **kwargs) -> None:  # noqa
    raise io.UnsupportedOperation("Binary file handlers can't be frozen")


def __freeze_object(
    obj: object, on_update: OnUpdateFuncType, on_freeze: OnFreezeFuncType, save_original_on_copy: bool = False
) -> FrozenBase:

    # If the object has a class with __slots__ a unique class is created whose class attributes
    # are the object attributes that we want to freeze
    if hasattr(obj.__class__, "__slots__"):
        attrs = tuple(obj.__class__.__slots__)
        on_freeze: OnFreezeFuncType = on_freeze_func_creator(on_freeze="copy")
        frozen_class = make_unique_class(
            klass=obj.__class__,
            attrs={
                attr: freeze(getattr(obj, attr), on_update=on_update, on_freeze=on_freeze, save_original_on_copy=False)
                for attr in attrs
            },
            on_update=on_update,
        )
        return frozen_class()
    else:
        attrs = tuple(obj.__dict__.keys())

        frozen_obj = on_freeze(obj)
        for attr in attrs:
            attr_value = getattr(frozen_obj, attr)
            setattr(
                frozen_obj,
                attr,
                freeze(attr_value, on_update=on_update, on_freeze=on_freeze, save_original_on_copy=False),
            )

        # Only when the frozen method is copying the objects we can get the original object
        # save_original_on_copy is used to save only the original object (the first-level object whose
        # save_original_on_copy is set to True). Descendant attributes are not saved in other original_obj attributes,
        # i.e. there is no copy of hierarchy, only the first-level object is saved.
        if save_original_on_copy and on_freeze.__class__ == OnFreezeCopier:
            setattr(frozen_obj, "original_obj", obj)

        frozen_class = make_frozen_class(klass=obj.__class__, attrs=attrs, on_update=on_update)
        frozen_obj.__class__ = frozen_class
        return frozen_obj


def __on_update_exception(frozen_obj: FrozenBase, message: str, *args, **kwargs) -> None:  # noqa
    raise FrozenException(message)


def __on_update_warning(frozen_obj: FrozenBase, message: str, *args, **kwargs) -> None:  # noqa
    warnings.warn(message)


def __on_update_func(on_update: OnUpdateFuncType) -> OnUpdateFuncType:
    if isinstance(on_update, str):
        if on_update == "exception":
            return __on_update_exception
        elif on_update == "warning":
            return __on_update_warning
        elif on_update == "nothing":
            return lambda message, *args, **kwargs: None
        else:
            raise AttributeError(
                f"Invalid value for on_update parameter, '{on_update}' found, "
                f"only 'exception', 'warning', and 'nothing' are valid options "
                f"if passed a string"
            )

    elif callable(on_update):
        return on_update

    else:
        raise AttributeError(
            f"Invalid value for on_update parameter, '{on_update}' found, "
            f"only 'exception', 'warning', 'nothing' or a function are "
            f"valid options"
        )



import unittest
from typing import Any, Iterator

from gelidum import FrozenException, freeze
from gelidum.collections.frozenlist import frozenlist
from gelidum.frozen import FrozenBase


class TestFrozenlist(unittest.TestCase):  # noqa
    def test_empty_construction(self):
        frozen_list = frozenlist()

        self.assertEqual(0, len(frozen_list))

    def test_construction(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_list = frozenlist([1, "2", Dummy(3)])

        self.assertTrue(isinstance(frozen_list, FrozenBase))
        self.assertEqual(3, len(frozen_list))
        self.assertEqual(1, frozen_list[0])
        self.assertEqual("2", frozen_list[1])
        self.assertTrue(isinstance(frozen_list[2], FrozenBase))
        self.assertEqual(3, frozen_list[2].value)

    def test_construction_from_tuple(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_list = frozenlist((1, "2", Dummy(3)))

        self.assertTrue(isinstance(frozen_list, FrozenBase))
        self.assertEqual(3, len(frozen_list))
        self.assertEqual(1, frozen_list[0])
        self.assertEqual("2", frozen_list[1])
        self.assertTrue(isinstance(frozen_list[2], FrozenBase))
        self.assertEqual(3, frozen_list[2].value)

    def test_construction_from_generator(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        def generator():
            yield 1
            yield "2"
            yield Dummy(3)

        frozen_list = frozenlist(generator())

        self.assertTrue(isinstance(frozen_list, FrozenBase))
        self.assertEqual(3, len(frozen_list))
        self.assertEqual(1, frozen_list[0])
        self.assertEqual("2", frozen_list[1])
        self.assertTrue(isinstance(frozen_list[2], FrozenBase))
        self.assertEqual(3, frozen_list[2].value)

    def test_setattr_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            setattr(frozen_list, "my_attr", "value")

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_count(self):
        frozen_list = frozenlist([1, 2, 3, 2, 3, 8, 3])

        self.assertEqual(1, frozen_list.count(1))
        self.assertEqual(2, frozen_list.count(2))
        self.assertEqual(3, frozen_list.count(3))
        self.assertEqual(1, frozen_list.count(8))
        self.assertEqual(0, frozen_list.count(10))

    def test_len(self):
        frozen_list = frozenlist([1, 2, 3])

        self.assertEqual(3, len(frozen_list))

    def test_add(self):
        frozen_list = frozenlist([1, 2, 3])
        frozen_list2 = frozen_list + frozenlist([2])

        self.assertTrue(isinstance(frozen_list, FrozenBase))
        self.assertEqual(3, len(frozen_list))
        self.assertEqual(1, frozen_list[0])
        self.assertEqual(2, frozen_list[1])
        self.assertEqual(3, frozen_list[2])
        self.assertTrue(isinstance(frozen_list2, FrozenBase))
        self.assertEqual(4, len(frozen_list2))
        self.assertEqual(1, frozen_list2[0])
        self.assertEqual(2, frozen_list2[1])
        self.assertEqual(3, frozen_list2[2])
        self.assertEqual(2, frozen_list2[3])

    def test_hash(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_list = frozenlist([1, 2, 3, [4, 5, 6], Dummy(2)])

        self.assertTrue(isinstance(hash(frozen_list), int))

    def test_mul(self):
        frozen_list = frozenlist([1, 2, 3])
        frozen_list2 = frozen_list * 2

        self.assertNotEqual(id(frozen_list), id(frozen_list2))
        self.assertTrue(isinstance(frozen_list2, FrozenBase))
        self.assertEqual(3, len(frozen_list))
        self.assertEqual(1, frozen_list[0])
        self.assertEqual(2, frozen_list[1])
        self.assertEqual(3, frozen_list[2])
        self.assertTrue(isinstance(frozen_list2, FrozenBase))
        self.assertEqual(6, len(frozen_list2))
        self.assertEqual(1, frozen_list2[0])
        self.assertEqual(2, frozen_list2[1])
        self.assertEqual(3, frozen_list2[2])
        self.assertEqual(1, frozen_list2[3])
        self.assertEqual(2, frozen_list2[4])
        self.assertEqual(3, frozen_list2[5])

    def test_slice(self):
        frozen_list = frozenlist([1, 2, 3])
        frozen_list_slice_2_items = frozen_list[0:2]
        frozen_list_slice_all_items = frozen_list[0:6]

        self.assertNotEqual(id(frozen_list), id(frozen_list_slice_2_items))
        self.assertNotEqual(id(frozen_list), id(frozen_list_slice_all_items))
        self.assertTrue(isinstance(frozen_list_slice_2_items, FrozenBase))
        self.assertEqual(2, len(frozen_list_slice_2_items))
        self.assertEqual(1, frozen_list_slice_2_items[0])
        self.assertEqual(2, frozen_list_slice_2_items[1])
        self.assertTrue(isinstance(frozen_list_slice_all_items, FrozenBase))
        self.assertEqual(3, len(frozen_list_slice_all_items))
        self.assertEqual(1, frozen_list_slice_all_items[0])
        self.assertEqual(2, frozen_list_slice_all_items[1])
        self.assertEqual(3, frozen_list_slice_all_items[2])

    def test_contains(self):
        frozen_list = frozenlist([1, 2, 3])

        self.assertTrue(1 in frozen_list)
        self.assertFalse(9 in frozen_list)

    def test_min(self):
        frozen_list = frozenlist([1, 2, 3])

        self.assertEqual(1, min(frozen_list))

    def test_max(self):
        frozen_list = frozenlist([1, 2, 3])

        self.assertEqual(3, max(frozen_list))

    def test_iter(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dummy = freeze(Dummy(9))
        frozen_list = frozenlist([1, 2, 3, frozen_dummy])
        items = []
        for item in frozen_list:
            items.append(item)

        self.assertListEqual([1, 2, 3, frozen_dummy], items)

    def test_next(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dummy = freeze(Dummy(9))
        frozen_list = frozenlist([1, 2, 3, frozen_dummy])
        frozen_list_iter = iter(frozen_list)

        self.assertEqual(1, next(frozen_list_iter))
        self.assertEqual(2, next(frozen_list_iter))
        self.assertEqual(3, next(frozen_list_iter))
        self.assertEqual(frozen_dummy, next(frozen_list_iter))

    def test_reversed(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_list = frozenlist([1, 2, 3, Dummy(9)])
        reversed_frozen_list = reversed(frozen_list)

        self.assertNotEqual(id(frozen_list), id(reversed_frozen_list))
        self.assertEqual(4, len(frozen_list))
        self.assertEqual(1, frozen_list[0])
        self.assertEqual(2, frozen_list[1])
        self.assertEqual(3, frozen_list[2])
        self.assertEqual(3, frozen_list[2])
        self.assertTrue(isinstance(frozen_list, FrozenBase))
        self.assertEqual(9, frozen_list[3].value)
        self.assertTrue(isinstance(reversed_frozen_list, Iterator))
        self.assertEqual(9, next(reversed_frozen_list).value)
        self.assertEqual(3, next(reversed_frozen_list))
        self.assertEqual(2, next(reversed_frozen_list))
        self.assertEqual(1, next(reversed_frozen_list))

    def test_copy(self):
        frozen_list = frozenlist([1, 2, 3])

        frozen_list_copy = frozen_list.copy()

        self.assertEqual(id(frozen_list_copy), id(frozen_list))

    def test_getitem(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(IndexError) as context:
            _val = frozen_list[4]  # noqa

        self.assertEqual(1, frozen_list[0])
        self.assertEqual(2, frozen_list[1])
        self.assertEqual(3, frozen_list[2])
        self.assertEqual("frozenlist index out of range", str(context.exception))

    def test_setitem_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_list[0] = 11

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_append_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_list.append(4)

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_extend_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_list.extend([4])

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_insert_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_list.insert(i=4, x=2)

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_remove_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_list.remove(4)

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_pop_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_list.pop(items=4)

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_clear_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_list.clear()

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_index_ok(self):
        frozen_list = frozenlist([1, 2, 3, 4, 5, 6, 7])

        index_for_2 = frozen_list.index(2)
        index_for_4_from_1_to_10 = frozen_list.index(4, 1, 10)

        with self.assertRaises(ValueError) as context_12:
            frozen_list.index(12)

        with self.assertRaises(ValueError) as context_2_from_5_to_10:
            frozen_list.index(2, 5, 10)

        self.assertEqual(1, index_for_2)
        self.assertEqual(3, index_for_4_from_1_to_10)
        self.assertEqual("12 is not in frozenlist", str(context_12.exception))
        self.assertEqual("2 is not in frozenlist", str(context_2_from_5_to_10.exception))

    def test_sort_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_list.sort()

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_reverse_exception(self):
        frozen_list = frozenlist([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_list.reverse()

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))



import unittest
from collections.abc import ValuesView, KeysView
from typing import Any, Tuple

from gelidum import FrozenException, freeze
from gelidum.collections.frozendict import frozendict
from gelidum.frozen import FrozenBase


class TestFrozendict(unittest.TestCase):  # noqa
    def test_empty_construction(self):
        frozen_list = frozendict()

        self.assertEqual(0, len(frozen_list))

    def test_construction_from_dict(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict({"one": 1, "two": "2", "three": Dummy(3)})

        self.assertTrue(isinstance(frozen_dict, FrozenBase))
        self.assertEqual(3, len(frozen_dict))
        self.assertEqual(1, frozen_dict["one"])
        self.assertEqual("2", frozen_dict["two"])
        self.assertTrue(isinstance(frozen_dict["three"], FrozenBase))
        self.assertEqual(3, frozen_dict["three"].value)

    def test_construction_from_empty_dict(self):
        frozen_dict = frozendict({})

        self.assertTrue(isinstance(frozen_dict, FrozenBase))
        self.assertEqual(0, len(frozen_dict))

    def test_construction_from_generator(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        def test_generator() -> Tuple[str, int]:
            yield "one", 1
            yield "two", "2"
            yield "three", Dummy(3)

        frozen_dict = frozendict(test_generator())

        self.assertTrue(isinstance(frozen_dict, FrozenBase))
        self.assertEqual(3, len(frozen_dict))
        self.assertEqual(1, frozen_dict["one"])
        self.assertEqual("2", frozen_dict["two"])
        self.assertTrue(isinstance(frozen_dict["three"], FrozenBase))
        self.assertEqual(3, frozen_dict["three"].value)

    def test_construction_from_kwargs(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict(one=1, two="2", three=Dummy(3))

        self.assertTrue(isinstance(frozen_dict, FrozenBase))
        self.assertEqual(3, len(frozen_dict))
        self.assertEqual(1, frozen_dict["one"])
        self.assertEqual("2", frozen_dict["two"])
        self.assertTrue(isinstance(frozen_dict["three"], FrozenBase))
        self.assertEqual(3, frozen_dict["three"].value)

    def test_construction_from_dict_and_kwargs(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict({"k0": "v0", "k1": "v1"}, one=1, two="2", three=Dummy(3))

        self.assertTrue(isinstance(frozen_dict, FrozenBase))
        self.assertEqual(5, len(frozen_dict))
        self.assertEqual("v0", frozen_dict["k0"])
        self.assertEqual("v1", frozen_dict["k1"])
        self.assertEqual(1, frozen_dict["one"])
        self.assertEqual("2", frozen_dict["two"])
        self.assertTrue(isinstance(frozen_dict["three"], FrozenBase))
        self.assertEqual(3, frozen_dict["three"].value)

    def test_construction_empty_dict_and_kwargs(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict({}, one=1, two="2", three=Dummy(3))

        self.assertTrue(isinstance(frozen_dict, FrozenBase))
        self.assertEqual(3, len(frozen_dict))
        self.assertEqual(1, frozen_dict["one"])
        self.assertEqual("2", frozen_dict["two"])
        self.assertTrue(isinstance(frozen_dict["three"], FrozenBase))
        self.assertEqual(3, frozen_dict["three"].value)

    def test_construction_from_list_of_pairs_and_kwargs(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict([("k0", "v0"), ("k1", "v1")], one=1, two="2", three=Dummy(3))

        self.assertTrue(isinstance(frozen_dict, FrozenBase))
        self.assertEqual(5, len(frozen_dict))
        self.assertEqual("v0", frozen_dict["k0"])
        self.assertEqual("v1", frozen_dict["k1"])
        self.assertEqual(1, frozen_dict["one"])
        self.assertEqual("2", frozen_dict["two"])
        self.assertTrue(isinstance(frozen_dict["three"], FrozenBase))
        self.assertEqual(3, frozen_dict["three"].value)

    def test_construction_list_of_strings_and_kwargs(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict(["k0", "v0", "k1", "v1"], one=1, two="2", three=Dummy(3))

        self.assertTrue(isinstance(frozen_dict, FrozenBase))
        self.assertEqual(5, len(frozen_dict))
        self.assertEqual("1", frozen_dict["k"])
        self.assertEqual("1", frozen_dict["v"])
        self.assertEqual(1, frozen_dict["one"])
        self.assertEqual("2", frozen_dict["two"])
        self.assertTrue(isinstance(frozen_dict["three"], FrozenBase))
        self.assertEqual(3, frozen_dict["three"].value)

    def test_construction_empty_list_and_kwargs(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict([], one=1, two="2", three=Dummy(3))

        self.assertTrue(isinstance(frozen_dict, FrozenBase))
        self.assertEqual(3, len(frozen_dict))
        self.assertEqual(1, frozen_dict["one"])
        self.assertEqual("2", frozen_dict["two"])
        self.assertTrue(isinstance(frozen_dict["three"], FrozenBase))
        self.assertEqual(3, frozen_dict["three"].value)

    def test_add(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dummy = freeze(Dummy(3))
        frozen_dict1 = frozendict({"one": 1, "two": "2", "three": frozen_dummy})
        frozen_dict2 = frozendict({"one": 1, "two": "2.5"})
        joined_frozen_dict = frozendict({"one": 1, "two": "2.5", "three": frozen_dummy})

        self.assertDictEqual(joined_frozen_dict, frozen_dict1 + frozen_dict2)

    def test_or(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dummy = freeze(Dummy(3))
        frozen_dict1 = frozendict({"one": 1, "two": "2", "three": frozen_dummy})
        frozen_dict2 = frozendict({"one": 1, "two": "2.5"})
        joined_frozen_dict = frozendict({"one": 1, "two": "2.5", "three": frozen_dummy})

        self.assertDictEqual(joined_frozen_dict, frozen_dict1 | frozen_dict2)

    def test_sub(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dummy = freeze(Dummy(3))
        frozen_dict1 = frozendict({"one": 1, "two": "2", "three": frozen_dummy})
        frozen_dict2 = frozendict({"one": 1, "two": "2.5"})
        joined_frozen_dict = frozendict({"three": frozen_dummy})

        self.assertDictEqual(joined_frozen_dict, frozen_dict1 - frozen_dict2)

    def test_getitem(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict({"one": 1, "two": "2", "three": Dummy(3)})

        self.assertEqual(1, frozen_dict["one"])
        self.assertEqual(1, frozen_dict.get("one"))
        self.assertEqual(None, frozen_dict.get("four"))
        self.assertEqual(4, frozen_dict.get("four", 4))

    def test_hash(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict({"one": 1, "two": "2", "three": Dummy(3), "four": {"a": "1"}})

        self.assertTrue(isinstance(hash(frozen_dict), int))

    def test_setitem(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict({"one": 1, "two": "2", "three": Dummy(3)})

        with self.assertRaises(FrozenException) as context_setitem:
            frozen_dict["four"] = 4

        self.assertEqual("'frozendict' object is immutable", str(context_setitem.exception))

    def test_del(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dict = frozendict({"one": 1, "two": "2", "three": Dummy(3)})
        with self.assertRaises(FrozenException) as context_del:
            del frozen_dict["one"]
        self.assertEqual("'frozendict' object is immutable", str(context_del.exception))

    def test_keys(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        dummy = Dummy(3)
        frozen_dict = frozendict({"one": 1, "two": "2", "three": dummy})
        frozen_dict_keys = frozen_dict.keys()
        frozen_dict_keys_iter = iter(frozen_dict_keys)
        frozen_dict_keys_list = list(frozen_dict_keys)

        self.assertTrue(isinstance(frozen_dict_keys, KeysView))
        self.assertEqual("one", next(frozen_dict_keys_iter))
        self.assertEqual("two", next(frozen_dict_keys_iter))
        self.assertEqual("three", next(frozen_dict_keys_iter))
        self.assertEqual(3, len(frozen_dict_keys_list))
        self.assertTrue("one" in frozen_dict_keys_list)
        self.assertTrue("two" in frozen_dict_keys_list)
        self.assertTrue("three" in frozen_dict_keys_list)

    def test_values(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        dummy = Dummy(3)
        frozen_dict = frozendict({"one": 1, "two": "2", "three": dummy})
        frozen_dict_values = frozen_dict.values()
        frozen_dict_values_iter = iter(frozen_dict.values())
        frozen_dict_values_list = list(frozen_dict_values)

        self.assertTrue(isinstance(frozen_dict_values, ValuesView))
        self.assertEqual(1, next(frozen_dict_values_iter))
        self.assertEqual("2", next(frozen_dict_values_iter))
        self.assertEqual(frozen_dict["three"], next(frozen_dict_values_iter))
        self.assertEqual(3, len(frozen_dict_values_list))
        self.assertTrue(1 in frozen_dict_values_list)
        self.assertTrue("2" in frozen_dict_values_list)
        self.assertFalse(dummy in frozen_dict_values_list)
        self.assertTrue(frozen_dict["three"] in frozen_dict_values_list)



import typing
from platform import python_implementation, python_version_tuple
from typing import Any, Callable, TYPE_CHECKING, Generic
from typing import Optional, Sized, Union, Iterable, Reversible, Mapping
from typing import TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from gelidum.frozen import FrozenBase  # noqa
    from gelidum.collections import frozendict, frozenlist, frozenzet, frozenndarray  # noqa


python_interpreter = python_implementation()
python_interpreter_version = tuple(int(number) for number in python_version_tuple()[:2])


if python_interpreter == "PyPy" or (python_interpreter_version[0] == 3 and python_interpreter_version[1] < 9):
    try:
        T = typing.TypeVar("T")

        class Final(Generic[T]):  # noqa
            pass

    except AttributeError:  # pragma: no cover
        Final = typing.Final

else:
    try:
        _SpecialForm = getattr(typing, "_SpecialForm")

        @_SpecialForm
        def Final(self, parameters):  # noqa
            return typing.Final[parameters]

    except AttributeError:  # pragma: no cover
        Final = typing.Final

FrozenList = Union["FrozenBase", Sized, Iterable, Reversible, "frozenlist"]
FrozenDict = Union["FrozenBase", Mapping, "frozendict"]
FrozenZet = Union["FrozenBase", Sized, Iterable, "frozenzet"]
FrozenNdArray = Union["FrozenBase", Sized, Iterable, "frozenndarray"]

T = TypeVar("T")

FrozenType = Optional[
    Union[
        bool,
        int,
        float,
        bytes,
        complex,
        str,
        bytes,
        FrozenDict,
        FrozenList,
        FrozenZet,
        FrozenNdArray,
        tuple,
        frozenset,
        "FrozenBase",
        T,
    ]
]

try:
    OnUpdateFuncType = Callable[["FrozenBase", str, ...], None]
except TypeError:
    OnUpdateFuncType = Callable


OnFreezeFuncType = Callable[[Any], Any]



class FrozenBase(object):
    @classmethod
    def _gelidum_on_update(cls, *args, **kwargs):  # pragma: no cover
        raise NotImplementedError("Implement in derived class")

    @classmethod
    def get_gelidum_hot_class_name(cls) -> str:  # pragma: no cover
        raise NotImplementedError("Implement in derived class")

    @classmethod
    def get_gelidum_hot_class_module(cls) -> str:  # pragma: no cover
        raise NotImplementedError("Implement in derived class")

    def __setattr__(self, key, value):
        self._gelidum_on_update(
            frozen_obj=self, message=f"Can't assign attribute '{key}' on immutable instance", key=key, value=value
        )

    def __set__(self, obj, value):
        self._gelidum_on_update(
            frozen_obj=self, message="Can't assign setter on immutable instance", obj=obj, value=value
        )

    def __delattr__(self, name):
        self._gelidum_on_update(
            frozen_obj=self, message=f"Can't delete attribute '{name}' on immutable instance", name=name
        )

    def __setitem__(self, key, value):
        self._gelidum_on_update(
            frozen_obj=self, message=f"Can't set key '{key}' on immutable instance", key=key, value=value
        )

    def __delitem__(self, key):
        self._gelidum_on_update(frozen_obj=self, message=f"Can't delete key '{key}' on immutable instance", key=key)

    def __deepcopy__(self, *args, **kwargs) -> "FrozenBase":
        """
        No frozen object must need to be (deep) copied.
        Use structural sharing to improve performance when doing
        immutable data structures.
        :return: reference to self.
        """
        return self



import copy
import datetime
import io
import json
import logging
import pickle
import sys
import tempfile
import threading
import unittest
import warnings
from typing import Dict, List, Union, Any
from unittest.mock import patch
from gelidum import FrozenException
from gelidum import freeze
from gelidum.dependencies import NUMPY_INSTALLED
from gelidum.collections import frozendict, frozenlist, frozenzet
from gelidum.frozen import FrozenBase, get_frozen_classes, clear_frozen_classes


class TestFreeze(unittest.TestCase):
    def setUp(self) -> None:
        clear_frozen_classes()

    def test_freeze_builtins(self):
        self.assertEqual(3, freeze(3))
        self.assertEqual(3.9, freeze(3.9))
        self.assertEqual(True, freeze(True))
        self.assertEqual(False, freeze(False))
        self.assertEqual(None, freeze(None))
        self.assertEqual(complex(1, 2), freeze(complex(1, 2)))
        self.assertEqual(b"Bytes", freeze(b"Bytes"))
        self.assertEqual("String", freeze("String"))

    def test_freeze_bytearray(self):
        self.assertEqual(b"Byte array", freeze(bytearray(b"Byte array")))

    def test_freeze_dict(self):
        frozen_obj: frozendict = freeze({"one": 1, "two": 2})

        with self.assertRaises(FrozenException) as context_assignment:
            frozen_obj["one"] = "another value"  # noqa

        with self.assertRaises(FrozenException) as context_clear:
            frozen_obj.clear()  # noqa

        with self.assertRaises(FrozenException) as context_update:
            frozen_obj.update({"three": 3})  # noqa

        with self.assertRaises(FrozenException) as context_deletion:
            del frozen_obj["one"]  # noqa

        self.assertEqual(2, len(frozen_obj))
        self.assertTrue("one" in frozen_obj)
        self.assertTrue("two" in frozen_obj)
        self.assertEqual(1, frozen_obj["one"])
        self.assertEqual(2, frozen_obj["two"])
        self.assertEqual("'frozendict' object is immutable", str(context_assignment.exception))
        self.assertEqual("'frozendict' object is immutable", str(context_clear.exception))
        self.assertEqual("'frozendict' object is immutable", str(context_update.exception))
        self.assertEqual("'frozendict' object is immutable", str(context_deletion.exception))

    def test_freeze_list(self):
        frozen_list = freeze(["one", 2, "three", ["a", "b", "c", 4, 5]])
        self.assertTrue(isinstance(frozen_list, frozenlist))
        self.assertEqual(4, len(frozen_list))
        self.assertEqual("one", frozen_list[0])
        self.assertEqual(2, frozen_list[1])
        self.assertEqual("three", frozen_list[2])
        self.assertTrue(isinstance(frozen_list[3], frozenlist))
        self.assertEqual(5, len(frozen_list[3]))
        self.assertEqual("a", frozen_list[3][0])
        self.assertEqual("b", frozen_list[3][1])
        self.assertEqual("c", frozen_list[3][2])
        self.assertEqual(4, frozen_list[3][3])
        self.assertEqual(5, frozen_list[3][4])

    def test_freeze_list_inplace_true_deprecated_parameter(self):
        with warnings.catch_warnings(record=True) as caught_warnings:
            frozen_list = freeze(["one", 2, "three"], inplace=True)

        self.assertTrue(isinstance(frozen_list, frozenlist))
        self.assertEqual(3, len(frozen_list))
        self.assertEqual("one", frozen_list[0])
        self.assertEqual(2, frozen_list[1])
        self.assertEqual("three", frozen_list[2])
        self.assertEqual(1, len(caught_warnings))
        self.assertEqual(
            "Use of inplace is deprecated and will be removed in next major version (0.6.0)",
            str(caught_warnings[0].message),
        )

    def test_freeze_tuple(self):
        self.assertEqual(("one", 2, "three"), freeze(("one", 2, "three")))

    def test_freeze_set(self):
        self.assertEqual(frozenzet(["one", 2, "three"]), freeze({"one", 2, "three"}))

    def test_freeze_simple_dataclass(self):
        from dataclasses import dataclass

        @dataclass
        class Dummy:
            attr1: str
            attr2: str
            attr3: str = "0"

        dummy = Dummy(attr1="1", attr2="2", attr3="3")
        frozen_dummy_on_freeze_copy = freeze(dummy, on_freeze="copy")
        frozen_dummy_inplace = freeze(dummy, on_freeze="inplace")

        with self.assertRaises(FrozenException) as context_on_freeze_copy:
            frozen_dummy_on_freeze_copy.attr1 = "2"

        with self.assertRaises(FrozenException) as context_inplace:
            frozen_dummy_inplace.attr2 = "2"

        self.assertEqual("Can't assign attribute 'attr1' on immutable instance", str(context_on_freeze_copy.exception))
        self.assertEqual("Can't assign attribute 'attr2' on immutable instance", str(context_inplace.exception))
        self.assertEqual(id(dummy), id(frozen_dummy_inplace))
        self.assertNotEqual(id(dummy), id(frozen_dummy_on_freeze_copy))

    def test_freeze_nested_dataclass(self):
        from dataclasses import dataclass

        @dataclass
        class ItemType:
            id: int
            name: str

        @dataclass
        class Item:
            id: int
            name: str
            order: int
            type: ItemType

        @dataclass
        class Database:
            items: List[Item]

        item_type1 = ItemType(1, "item type 1")
        item_type2 = ItemType(2, "item type 2")

        dummy_database = Database(
            items=[
                Item(1, "item 1", 5, item_type1),
                Item(2, "item 2", 1, item_type1),
                Item(3, "item 3", 99, item_type2),
            ]
        )
        frozen_dummy_on_freeze_copy = freeze(dummy_database, on_freeze="copy")
        frozen_dummy_inplace = freeze(dummy_database, on_freeze="inplace")

        with self.assertRaises(FrozenException) as context_on_freeze_copy_items_assignment:
            frozen_dummy_on_freeze_copy.items = []

        with self.assertRaises(FrozenException) as context_on_freeze_copy_item_assignment:
            frozen_dummy_on_freeze_copy.items[0] = Item(9, "item 9", 99, item_type1)

        with self.assertRaises(FrozenException) as context_inplace:
            frozen_dummy_inplace.attr2 = "2"

        self.assertEqual(
            "Can't assign attribute 'items' on immutable instance",
            str(context_on_freeze_copy_items_assignment.exception),
        )
        self.assertEqual("'frozenlist' object is immutable", str(context_on_freeze_copy_item_assignment.exception))
        self.assertEqual("Can't assign attribute 'attr2' on immutable instance", str(context_inplace.exception))
        self.assertEqual(id(dummy_database), id(frozen_dummy_inplace))
        self.assertNotEqual(id(dummy_database), id(frozen_dummy_on_freeze_copy))

    def test_freeze_objects_by_copying_and_access_original_object(self):
        class DummyChild(object):
            def __init__(self, value: int):
                self.attr = value

        class Dummy(object):
            def __init__(self, child: DummyChild):
                self.child = child

        dummy_child = DummyChild(value=1)
        dummy = Dummy(child=dummy_child)
        frozen_dummy = freeze(dummy, on_freeze="copy", save_original_on_copy=True)

        self.assertNotEqual(frozen_dummy.__class__, dummy.__class__)
        self.assertEqual(frozen_dummy.original_obj.__class__, dummy.__class__)
        self.assertEqual(id(frozen_dummy.original_obj), id(dummy))
        self.assertEqual(frozen_dummy.child.original_obj, None)

    def test_freeze_objects_by_copying_but_not_saving_original_object(self):
        class DummyChild(object):
            def __init__(self, value: int):
                self.attr = value

        class Dummy(object):
            def __init__(self, child: DummyChild):
                self.child = child

        dummy_child = DummyChild(value=1)
        dummy = Dummy(child=dummy_child)
        frozen_dummy = freeze(dummy, on_freeze="copy", save_original_on_copy=False)

        self.assertNotEqual(frozen_dummy.__class__, dummy.__class__)
        self.assertIsNone(frozen_dummy.original_obj)

    def test_freeze_objects_inplace_and_no_access_to_original_object(self):
        class Dummy(object):
            def __init__(self, value: int):
                self.attr = value

        dummy = Dummy(value=1)
        frozen_dummy = freeze(dummy, on_freeze="inplace")

        self.assertEqual(frozen_dummy.__class__, dummy.__class__)
        self.assertIsNone(frozen_dummy.original_obj)

    def test_freeze_objects_of_same_class(self):
        class Dummy(object):
            def __init__(self, value: int):
                self.attr = value

        dummy1 = Dummy(value=1)
        frozen_dummy1 = freeze(dummy1, on_freeze="copy")
        dummy2 = Dummy(value=2)
        frozen_dummy2 = freeze(dummy2, on_freeze="copy")

        self.assertEqual(frozen_dummy1.__class__, frozen_dummy2.__class__)

    def test_freeze_simple_object_inplace_true_deprecated_parameter(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        with warnings.catch_warnings(record=True) as caught_warnings:
            frozen_dummy = freeze(dummy, on_update="exception", inplace=True)

        self.assertEqual(id(dummy), id(frozen_dummy))
        self.assertEqual(1, len(caught_warnings))
        self.assertEqual(
            "Use of inplace is deprecated and will be removed in next major version (0.6.0)",
            str(caught_warnings[0].message),
        )

    def test_freeze_simple_object_inplace_false_deprecated_parameter(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        with warnings.catch_warnings(record=True) as caught_warnings:
            frozen_dummy = freeze(dummy, on_update="exception", inplace=False)

        self.assertNotEqual(id(dummy), id(frozen_dummy))
        self.assertEqual(1, len(caught_warnings))
        self.assertEqual(
            "Use of inplace is deprecated and will be removed in next major version (0.6.0)",
            str(caught_warnings[0].message),
        )

    def test_freeze_simple_object_on_update_warning_on_freeze_inplace(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy, on_update="warning", on_freeze="inplace")

        with warnings.catch_warnings(record=True) as caught_warnings:
            frozen_dummy.attr1 = 99
            frozen_dummy._attr2 = 99
            frozen_dummy._Dummy__attr3 = 99

        self.assertEqual(id(dummy), id(frozen_dummy))
        self.assertEqual((FrozenBase, Dummy), frozen_dummy.__class__.__bases__)
        self.assertEqual(1, frozen_dummy.attr1)
        self.assertEqual(2, frozen_dummy._attr2)
        self.assertEqual(3, frozen_dummy._Dummy__attr3)
        self.assertListEqual(
            [
                "Can't assign attribute 'attr1' on immutable instance",
                "Can't assign attribute '_attr2' on immutable instance",
                "Can't assign attribute '_Dummy__attr3' on immutable instance",
            ],
            [str(warn.message) for warn in caught_warnings],
        )

    def test_freeze_object_of_class_with_slots(self):
        class ClassWithSlots(object):
            __slots__ = ("attr1", "_attr2")

            def __init__(self, attr1: int, attr2: int):
                self.attr1 = attr1
                self._attr2 = attr2

        obj = ClassWithSlots(attr1=1, attr2=2)
        frozen_obj = freeze(obj, on_update="warning", on_freeze="copy")

        with warnings.catch_warnings(record=True) as caught_warnings:
            frozen_obj.attr1 = 99
            frozen_obj._attr2 = 99
            frozen_obj.new_attribute = 99

        self.assertNotEqual(id(obj), id(frozen_obj))
        self.assertEqual((FrozenBase, ClassWithSlots), frozen_obj.__class__.__bases__)
        self.assertEqual(1, frozen_obj.attr1)
        self.assertEqual(2, frozen_obj._attr2)
        self.assertListEqual(
            [
                "Can't assign attribute 'attr1' on immutable instance",
                "Can't assign attribute '_attr2' on immutable instance",
                "Can't assign attribute 'new_attribute' on immutable instance",
            ],
            [str(warn.message) for warn in caught_warnings],
        )

    def test_freeze_two_object_of_the_same_class_with_slots(self):
        class ClassWithSlots(object):
            __slots__ = ("attr1", "_attr2")

            def __init__(self, attr1: int, attr2: int):
                self.attr1 = attr1
                self._attr2 = attr2

        obj1 = ClassWithSlots(attr1=1, attr2=2)
        frozen_obj1 = freeze(obj1, on_update="warning", on_freeze="copy")

        obj2 = ClassWithSlots(attr1=3, attr2=4)
        frozen_obj2 = freeze(obj2, on_update="warning", on_freeze="copy")

        self.assertNotEqual(id(frozen_obj1.__class__), id(frozen_obj2.__class__))
        self.assertEqual((FrozenBase, ClassWithSlots), frozen_obj1.__class__.__bases__)
        self.assertEqual((FrozenBase, ClassWithSlots), frozen_obj2.__class__.__bases__)
        self.assertEqual(1, frozen_obj1.attr1)
        self.assertEqual(2, frozen_obj1._attr2)
        self.assertEqual(3, frozen_obj2.attr1)
        self.assertEqual(4, frozen_obj2._attr2)

    def test_freeze_object_with_class_with_slots_whose_attributes_are_objects_with_class_with_slots(self):
        class Attr(object):
            __slots__ = ("value",)

            def __init__(self, value: int):
                self.value = value

        class ClassWithSlots(object):
            __slots__ = ("attr1", "attr2")

            def __init__(self, attr1: Attr, attr2: Attr):
                self.attr1 = attr1
                self.attr2 = attr2

        obj = ClassWithSlots(attr1=Attr(1), attr2=Attr(2))
        frozen_obj = freeze(obj, on_update="warning", on_freeze="copy")

        with warnings.catch_warnings(record=True) as caught_warnings:
            frozen_obj.attr1 = 99
            frozen_obj.attr2 = 99

        self.assertNotEqual(id(obj), id(frozen_obj))
        self.assertEqual((FrozenBase, ClassWithSlots), frozen_obj.__class__.__bases__)
        self.assertEqual(1, frozen_obj.attr1.value)
        self.assertEqual(2, frozen_obj.attr2.value)
        self.assertListEqual(
            [
                "Can't assign attribute 'attr1' on immutable instance",
                "Can't assign attribute 'attr2' on immutable instance",
            ],
            [str(warn.message) for warn in caught_warnings],
        )

    def test_freeze_object_of_class_with_slots_inplace(self):
        class ClassWithSlots(object):
            __slots__ = ("attr1", "_attr2")

            def __init__(self, attr1: int, attr2: int):
                self.attr1 = attr1
                self._attr2 = attr2

        obj = ClassWithSlots(attr1=1, attr2=2)
        with self.assertRaises(FrozenException) as freeze_context:
            freeze(obj, on_update="warning", on_freeze="inplace")

        self.assertEqual("Objects of classes with __slots__ cannot be frozen inplace", str(freeze_context.exception))

    def test_freeze_object_of_class_with_slots_inplace_deprecated_parameter(self):
        class ClassWithSlots(object):
            __slots__ = ("attr1", "_attr2")

            def __init__(self, attr1: int, attr2: int):
                self.attr1 = attr1
                self._attr2 = attr2

        obj = ClassWithSlots(attr1=1, attr2=2)
        with self.assertRaises(FrozenException) as freeze_context:
            freeze(obj, on_update="warning", inplace=True)

        self.assertEqual("Objects of classes with __slots__ cannot be frozen inplace", str(freeze_context.exception))

    def test_freeze_simple_object_on_update_nothing_on_freeze_inplace(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy, on_update="nothing", on_freeze="inplace")
        frozen_dummy.attr1 = 99
        frozen_dummy._attr2 = 99
        frozen_dummy._Dummy__attr3 = 99

        self.assertEqual(id(dummy), id(frozen_dummy))
        self.assertEqual((FrozenBase, Dummy), frozen_dummy.__class__.__bases__)
        self.assertEqual(1, frozen_dummy.attr1)
        self.assertEqual(2, frozen_dummy._attr2)
        self.assertEqual(3, frozen_dummy._Dummy__attr3)

    def test_freeze_simple_object_and_catch_setattr_exception(self):
        class Dummy(object):
            def __init__(self, value: int):
                self.attr = value

        class DummyWithProperty(object):
            def __init__(self, value: int):
                self.__attr = value

            @property
            def attr(self):
                return self.__attr

        dummy = Dummy(value=1)
        frozen_dummy = freeze(dummy, on_update="exception", on_freeze="inplace")

        dummy_with_property = DummyWithProperty(value=1)
        frozen_dummy_with_property = freeze(dummy_with_property, on_update="exception", on_freeze="inplace")

        with self.assertRaises(FrozenException) as setattr_context:
            setattr(frozen_dummy, "attr", 99)

        with self.assertRaises(FrozenException) as setattr_context2:
            frozen_dummy_with_property.attr = "99"

        with self.assertRaises(FrozenException) as delattr_context:
            delattr(frozen_dummy, "attr")

        with self.assertRaises(FrozenException) as setitem_context:
            frozen_dummy["attr"] = 99

        with self.assertRaises(FrozenException) as delitem_context:
            del frozen_dummy["attr"]

        self.assertEqual(id(dummy), id(frozen_dummy))
        self.assertEqual((FrozenBase, Dummy), frozen_dummy.__class__.__bases__)
        self.assertEqual(1, frozen_dummy.attr)
        self.assertEqual(1, frozen_dummy_with_property.attr)
        self.assertEqual("Can't assign attribute 'attr' on immutable instance", str(setattr_context.exception))
        self.assertEqual("Can't assign attribute 'attr' on immutable instance", str(setattr_context2.exception))
        self.assertEqual("Can't delete attribute 'attr' on immutable instance", str(delattr_context.exception))
        self.assertEqual("Can't set key 'attr' on immutable instance", str(setitem_context.exception))
        self.assertEqual("Can't delete key 'attr' on immutable instance", str(delitem_context.exception))

    def test_freeze_object_with_set_descriptor_method_exception(self):
        class DummyCounter(object):
            def __init__(self, name: str):
                self.name = name

            def __get__(self, obj, type=None) -> object:  # noqa
                return obj.__dict__.get(self.name) or 0  # pragma: no cover

            def __set__(self, obj, value) -> None:
                obj.__dict__[self.name] = value  # pragma: no cover

        class Dummy:
            counter = freeze(DummyCounter("counter"))

            def __init__(self, value: int):
                self.attr = value

        dummy1 = Dummy(value=1)

        with self.assertRaises(FrozenException) as set_context:
            dummy1.counter = 1

        self.assertEqual("Can't assign setter on immutable instance", str(set_context.exception))

    @patch("datetime.datetime")
    def test_freeze_simple_object_on_update_function_freeze_inplace(self, mock_datetime):
        fixed_utcnow = datetime.datetime(2021, 6, 16, 14, 20, 56, 809581)
        mock_datetime.utcnow = unittest.mock.Mock(return_value=fixed_utcnow)

        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        writing_tries = []

        def on_update_func(message, *args, **kwargs):
            writing_tries.append(
                {"message": message, "args": args, "kwargs": kwargs, "time": datetime.datetime.utcnow()}
            )

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy, on_update=on_update_func, on_freeze="inplace")
        frozen_dummy.attr1 = 99
        frozen_dummy._attr2 = 99
        frozen_dummy._Dummy__attr3 = 99

        expected_writing_tries = [
            {
                "args": (),
                "kwargs": {"key": "attr1", "value": 99, "frozen_obj": frozen_dummy},
                "message": "Can't assign attribute 'attr1' on immutable instance",
                "time": fixed_utcnow,
            },
            {
                "args": (),
                "kwargs": {"key": "_attr2", "value": 99, "frozen_obj": frozen_dummy},
                "message": "Can't assign attribute '_attr2' on immutable instance",
                "time": fixed_utcnow,
            },
            {
                "args": (),
                "kwargs": {"key": "_Dummy__attr3", "value": 99, "frozen_obj": frozen_dummy},
                "message": "Can't assign attribute '_Dummy__attr3' on immutable instance",
                "time": fixed_utcnow,
            },
        ]

        self.assertEqual(id(dummy), id(frozen_dummy))
        self.assertEqual((FrozenBase, Dummy), frozen_dummy.__class__.__bases__)
        self.assertEqual(1, frozen_dummy.attr1)
        self.assertEqual(2, frozen_dummy._attr2)
        self.assertEqual(3, frozen_dummy._Dummy__attr3)
        self.assertListEqual(expected_writing_tries, writing_tries)

    @patch("logging.Logger")
    def test_freeze_simple_object_on_update_func_store_update_tries(self, mock_logger):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self.attr2 = attr2
                self.attr3 = attr3

        class FrozenDummyUpdateTryRecorder(object):
            def __init__(self, log: logging.Logger):
                self.log = log
                self.lock = threading.Lock()
                self.writing_tries: List[Dict] = []
                self.original_obj = None

            def on_freeze(self, obj: object) -> object:
                frozen_object = copy.deepcopy(obj)
                self.original_obj = obj
                return frozen_object

            def on_update(self, frozen_obj: "FrozenBase", message: str, *args, **kwargs):
                self.log.warning(message)
                with self.lock:
                    self.writing_tries.append(
                        {
                            "frozen_obj": frozen_obj,
                            "args": args,
                            "kwargs": kwargs,
                            "datetime": datetime.datetime.utcnow(),
                        }
                    )

        update_try_recorder = FrozenDummyUpdateTryRecorder(log=mock_logger)

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy, on_freeze=update_try_recorder.on_freeze, on_update=update_try_recorder.on_update)

        expected_writing_tries = [
            {
                "args": tuple(),
                "datetime": unittest.mock.ANY,
                "frozen_obj": frozen_dummy,
                "kwargs": {"key": "attr1", "value": 99},
            }
        ]

        frozen_dummy.attr1 = 99

        self.assertNotEqual(id(dummy), id(frozen_dummy))
        self.assertEqual((FrozenBase, Dummy), frozen_dummy.__class__.__bases__)
        self.assertEqual(1, frozen_dummy.attr1)
        self.assertEqual(2, frozen_dummy.attr2)
        self.assertEqual(3, frozen_dummy.attr3)
        self.assertEqual(
            unittest.mock.call("Can't assign attribute 'attr1' on immutable instance"), mock_logger.warning.call_args
        )
        self.assertEqual(id(dummy), id(update_try_recorder.original_obj))
        self.assertEqual(expected_writing_tries, update_try_recorder.writing_tries)

    def test_freeze_simple_object_on_freeze_inplace(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy, on_freeze="inplace")

        with self.assertRaises(FrozenException) as context_exc_assign_attr1:
            frozen_dummy.attr1 = 99
        with self.assertRaises(FrozenException) as context_exc_assign_attr2:
            frozen_dummy._attr2 = 99
        with self.assertRaises(FrozenException) as context_exc_assign_attr3:
            frozen_dummy._Dummy__attr3 = 99
        with self.assertRaises(FrozenException) as context_exc_inc_attr1:
            frozen_dummy.attr1 += 99
        with self.assertRaises(FrozenException) as context_exc_inc_attr2:
            frozen_dummy._attr2 += 99
        with self.assertRaises(FrozenException) as context_exc_inc_attr3:
            frozen_dummy._Dummy__attr3 += 99
        with self.assertRaises(FrozenException) as context_exc_dec_attr1:
            frozen_dummy.attr1 -= 99
        with self.assertRaises(FrozenException) as context_exc_dec_attr2:
            frozen_dummy._attr2 -= 99
        with self.assertRaises(FrozenException) as context_exc_dec_attr3:
            frozen_dummy._Dummy__attr3 -= 99

        self.assertEqual(id(dummy), id(frozen_dummy))
        self.assertEqual((FrozenBase, Dummy), frozen_dummy.__class__.__bases__)
        self.assertEqual(1, frozen_dummy.attr1)
        self.assertEqual(2, frozen_dummy._attr2)
        self.assertEqual(3, frozen_dummy._Dummy__attr3)
        self.assertEqual(
            "Can't assign attribute 'attr1' on immutable instance", str(context_exc_assign_attr1.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_attr2' on immutable instance", str(context_exc_assign_attr2.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance", str(context_exc_assign_attr3.exception)
        )
        self.assertEqual("Can't assign attribute 'attr1' on immutable instance", str(context_exc_inc_attr1.exception))
        self.assertEqual("Can't assign attribute '_attr2' on immutable instance", str(context_exc_inc_attr2.exception))
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance", str(context_exc_inc_attr3.exception)
        )
        self.assertEqual("Can't assign attribute 'attr1' on immutable instance", str(context_exc_dec_attr1.exception))
        self.assertEqual("Can't assign attribute '_attr2' on immutable instance", str(context_exc_dec_attr2.exception))
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance", str(context_exc_dec_attr3.exception)
        )

    def test_freeze_simple_object_on_freeze_copy(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy, on_freeze="copy")

        with self.assertRaises(FrozenException) as context_exc_assign_attr1:
            frozen_dummy.attr1 = 99
        with self.assertRaises(FrozenException) as context_exc_assign_attr2:
            frozen_dummy._attr2 = 99
        with self.assertRaises(FrozenException) as context_exc_assign_attr3:
            frozen_dummy._Dummy__attr3 = 99
        with self.assertRaises(FrozenException) as context_exc_inc_attr1:
            frozen_dummy.attr1 += 99
        with self.assertRaises(FrozenException) as context_exc_inc_attr2:
            frozen_dummy._attr2 += 99
        with self.assertRaises(FrozenException) as context_exc_inc_attr3:
            frozen_dummy._Dummy__attr3 += 99
        with self.assertRaises(FrozenException) as context_exc_dec_attr1:
            frozen_dummy.attr1 -= 99
        with self.assertRaises(FrozenException) as context_exc_dec_attr2:
            frozen_dummy._attr2 -= 99
        with self.assertRaises(FrozenException) as context_exc_dec_attr3:
            frozen_dummy._Dummy__attr3 -= 99

        self.assertNotEqual(id(dummy), id(frozen_dummy))
        self.assertEqual((FrozenBase, Dummy), frozen_dummy.__class__.__bases__)
        self.assertEqual(1, frozen_dummy.attr1)
        self.assertEqual(2, frozen_dummy._attr2)
        self.assertEqual(3, frozen_dummy._Dummy__attr3)
        self.assertEqual(
            "Can't assign attribute 'attr1' on immutable instance", str(context_exc_assign_attr1.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_attr2' on immutable instance", str(context_exc_assign_attr2.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance", str(context_exc_assign_attr3.exception)
        )
        self.assertEqual("Can't assign attribute 'attr1' on immutable instance", str(context_exc_inc_attr1.exception))
        self.assertEqual("Can't assign attribute '_attr2' on immutable instance", str(context_exc_inc_attr2.exception))
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance", str(context_exc_inc_attr3.exception)
        )
        self.assertEqual("Can't assign attribute 'attr1' on immutable instance", str(context_exc_dec_attr1.exception))
        self.assertEqual("Can't assign attribute '_attr2' on immutable instance", str(context_exc_dec_attr2.exception))
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance", str(context_exc_dec_attr3.exception)
        )

    def test_freeze_deep_object_on_freeze_inplace(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        class DeepDummy(object):
            def __init__(self, dummy1: Dummy, dummy2: Dummy, dummy3: Dummy):
                self.dummy1 = dummy1
                self._dummy2 = dummy2
                self.__dummy3 = dummy3

        deep_dummy = DeepDummy(
            dummy1=Dummy(attr1=11, attr2=12, attr3=13),
            dummy2=Dummy(attr1=21, attr2=22, attr3=23),
            dummy3=Dummy(attr1=31, attr2=32, attr3=33),
        )

        frozen_deep_dummy = freeze(deep_dummy, on_freeze="inplace")

        with self.assertRaises(FrozenException) as context_exc_dummy1:
            frozen_deep_dummy.dummy1 = 99
        with self.assertRaises(FrozenException) as context_exc_dummy1_attr1:
            frozen_deep_dummy.dummy1.attr1 = 99
        with self.assertRaises(FrozenException) as context_exc_dummy1_attr2:
            frozen_deep_dummy.dummy1._attr2 = 99
        with self.assertRaises(FrozenException) as context_exc_dummy1_attr3:
            frozen_deep_dummy.dummy1._Dummy__attr3 = 99
        with self.assertRaises(FrozenException) as context_exc_setattr_dummy1:
            setattr(frozen_deep_dummy, "dummy1", 99)
        with self.assertRaises(FrozenException) as context_exc_setattr_dummy1_attr1:
            setattr(frozen_deep_dummy.dummy1, "attr1", 99)
        with self.assertRaises(FrozenException) as context_exc_setattr_dummy1_attr2:
            setattr(frozen_deep_dummy.dummy1, "_attr2", 99)
        with self.assertRaises(FrozenException) as context_exc_setattr_dummy1_attr3:
            setattr(frozen_deep_dummy.dummy1, "_Dummy__attr3", 99)

        self.assertEqual(id(deep_dummy), id(frozen_deep_dummy))
        self.assertEqual(11, frozen_deep_dummy.dummy1.attr1)
        self.assertEqual(12, frozen_deep_dummy.dummy1._attr2)
        self.assertEqual(13, frozen_deep_dummy.dummy1._Dummy__attr3)
        self.assertEqual("Can't assign attribute 'dummy1' on immutable instance", str(context_exc_dummy1.exception))
        self.assertEqual(
            "Can't assign attribute 'attr1' on immutable instance", str(context_exc_dummy1_attr1.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_attr2' on immutable instance", str(context_exc_dummy1_attr2.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance", str(context_exc_dummy1_attr3.exception)
        )
        self.assertEqual(
            "Can't assign attribute 'dummy1' on immutable instance", str(context_exc_setattr_dummy1.exception)
        )
        self.assertEqual(
            "Can't assign attribute 'attr1' on immutable instance", str(context_exc_setattr_dummy1_attr1.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_attr2' on immutable instance", str(context_exc_setattr_dummy1_attr2.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance",
            str(context_exc_setattr_dummy1_attr3.exception),
        )

    def test_freeze_deep_object_on_freeze_copy(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        class DeepDummy(object):
            def __init__(self, dummy1: Dummy, dummy2: Dummy, dummy3: Dummy):
                self.dummy1 = dummy1
                self._dummy2 = dummy2
                self.__dummy3 = dummy3

        deep_dummy = DeepDummy(
            dummy1=Dummy(attr1=11, attr2=12, attr3=13),
            dummy2=Dummy(attr1=21, attr2=22, attr3=23),
            dummy3=Dummy(attr1=31, attr2=32, attr3=33),
        )

        frozen_deep_dummy = freeze(deep_dummy, on_freeze="copy")

        with self.assertRaises(FrozenException) as context_exc_dummy1:
            frozen_deep_dummy.dummy1 = 99
        with self.assertRaises(FrozenException) as context_exc_dummy1_attr1:
            frozen_deep_dummy.dummy1.attr1 = 99
        with self.assertRaises(FrozenException) as context_exc_dummy1_attr2:
            frozen_deep_dummy.dummy1._attr2 = 99
        with self.assertRaises(FrozenException) as context_exc_dummy1_attr3:
            frozen_deep_dummy.dummy1._Dummy__attr3 = 99
        with self.assertRaises(FrozenException) as context_exc_setattr_dummy1:
            setattr(frozen_deep_dummy, "dummy1", 99)
        with self.assertRaises(FrozenException) as context_exc_setattr_dummy1_attr1:
            setattr(frozen_deep_dummy.dummy1, "attr1", 99)
        with self.assertRaises(FrozenException) as context_exc_setattr_dummy1_attr2:
            setattr(frozen_deep_dummy.dummy1, "_attr2", 99)
        with self.assertRaises(FrozenException) as context_exc_setattr_dummy1_attr3:
            setattr(frozen_deep_dummy.dummy1, "_Dummy__attr3", 99)

        self.assertNotEqual(id(deep_dummy), id(frozen_deep_dummy))
        self.assertEqual(11, frozen_deep_dummy.dummy1.attr1)
        self.assertEqual(12, frozen_deep_dummy.dummy1._attr2)
        self.assertEqual(13, frozen_deep_dummy.dummy1._Dummy__attr3)
        self.assertEqual("Can't assign attribute 'dummy1' on immutable instance", str(context_exc_dummy1.exception))
        self.assertEqual(
            "Can't assign attribute 'attr1' on immutable instance", str(context_exc_dummy1_attr1.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_attr2' on immutable instance", str(context_exc_dummy1_attr2.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance", str(context_exc_dummy1_attr3.exception)
        )
        self.assertEqual(
            "Can't assign attribute 'dummy1' on immutable instance", str(context_exc_setattr_dummy1.exception)
        )
        self.assertEqual(
            "Can't assign attribute 'attr1' on immutable instance", str(context_exc_setattr_dummy1_attr1.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_attr2' on immutable instance", str(context_exc_setattr_dummy1_attr2.exception)
        )
        self.assertEqual(
            "Can't assign attribute '_Dummy__attr3' on immutable instance",
            str(context_exc_setattr_dummy1_attr3.exception),
        )

    def test_freeze_object_with_file_handler_attribute(self):
        with tempfile.TemporaryFile("w") as temp_text_file:
            with tempfile.TemporaryFile("wb") as temp_bin_file:

                class DummyWithTextFile(object):
                    def __init__(self):
                        self.text_file = temp_text_file

                class DummyWithBinaryFile(object):
                    def __init__(self):
                        self.binary_file = temp_bin_file

                dummy_with_text_file = DummyWithTextFile()
                dummy_with_binary_file = DummyWithBinaryFile()

                with self.assertRaises(io.UnsupportedOperation) as text_write_context:
                    freeze(dummy_with_text_file, on_freeze="inplace")

                with self.assertRaises(io.UnsupportedOperation) as binary_write_context:
                    freeze(dummy_with_binary_file, on_freeze="inplace")

        self.assertEqual("Text file handlers can't be frozen", str(text_write_context.exception))
        self.assertEqual("Binary file handlers can't be frozen", str(binary_write_context.exception))

    def test_hash_on_freeze_inplace(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy, on_freeze="inplace")
        self.assertEqual(hash(dummy), hash(frozen_dummy))

    def test_hash_on_freeze_copy(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy, on_freeze="copy")
        self.assertNotEqual(hash(dummy), hash(frozen_dummy))

    def test_hash_frozen_object_as_key_in_dict(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy, on_freeze="copy")
        frozen_dummy_inplace = freeze(dummy, on_freeze="inplace")
        my_dict: Dict = {dummy: "this is a dummy object"}

        self.assertNotEqual(hash(dummy), hash(frozen_dummy))
        self.assertEqual(None, my_dict.get(frozen_dummy))
        self.assertEqual(hash(dummy), hash(frozen_dummy_inplace))
        self.assertEqual("this is a dummy object", my_dict.get(frozen_dummy_inplace))

    def test_modifying_class_attribute_on_frozen_objects(self):
        class Dummy(object):
            COUNTER = 1

        dummy = Dummy()
        frozen_dummy = freeze(dummy, on_freeze="copy")
        frozen_dummy_inplace = freeze(dummy, on_freeze="inplace")
        Dummy.COUNTER += 1

        self.assertEqual(2, Dummy.COUNTER)
        self.assertEqual(2, frozen_dummy.COUNTER)
        self.assertEqual(2, frozen_dummy_inplace.COUNTER)

    def test_cannot_add_attribute_to_frozen_object(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy)
        with self.assertRaises(FrozenException) as context:
            frozen_dummy.new_attribute = 99
        self.assertEqual("Can't assign attribute 'new_attribute' on immutable instance", str(context.exception))

    def test_cannot_use_modifying_attribute_method_in_frozen_object(self):
        class Dummy(object):
            def __init__(self, attr1: int, attr2: int, attr3: int):
                self.attr1 = attr1
                self._attr2 = attr2
                self.__attr3 = attr3

            def inc_attr1(self):
                self.attr1 += 1

            def inc_attr2(self):
                self._attr2 += 1

            def inc_attr3(self):
                self.__attr3 += 1

        dummy = Dummy(attr1=1, attr2=2, attr3=3)
        frozen_dummy = freeze(dummy)
        with self.assertRaises(FrozenException) as context1:
            frozen_dummy.inc_attr1()
        with self.assertRaises(FrozenException) as context2:
            frozen_dummy.inc_attr2()
        with self.assertRaises(FrozenException) as context3:
            frozen_dummy.inc_attr3()

        self.assertEqual("Can't assign attribute 'attr1' on immutable instance", str(context1.exception))
        self.assertEqual("Can't assign attribute '_attr2' on immutable instance", str(context2.exception))
        self.assertEqual("Can't assign attribute '_Dummy__attr3' on immutable instance", str(context3.exception))

    def test_json(self):
        data_list = [{"a": 1, "b": 2}, {"a": 11, "b": 22}]
        data_dict = {"first": [{"a": 1}, {"a": 1}]}

        self.assertEqual('[{"a": 1, "b": 2}, {"a": 11, "b": 22}]', json.dumps(freeze(data_list)))
        self.assertEqual('{"first": [{"a": 1}, {"a": 1}]}', json.dumps(freeze(data_dict)))

    def test_pickle(self):
        class DummyForPickle(object):
            def __init__(self, attr: int):
                self.attr = attr

        setattr(sys.modules[__name__], DummyForPickle.__name__, DummyForPickle)

        dummy = DummyForPickle(attr=1)
        frozen_dummy = freeze(dummy)
        pickled_frozen_dummy = pickle.dumps(frozen_dummy)
        unpickled_frozen_dummy = pickle.loads(pickled_frozen_dummy)

        with self.assertRaises(FrozenException) as context:
            unpickled_frozen_dummy.attr = 8

        self.assertEqual("Can't assign attribute 'attr' on immutable instance", str(context.exception))
        self.assertEqual(1, frozen_dummy.attr, unpickled_frozen_dummy.attr)

    def test_pickle_object_of_same_name_classes(self):
        import gelidum.tests.gelidum_tests.utils.dummy1
        import gelidum.tests.gelidum_tests.utils.dummy2

        dummy1 = gelidum.tests.gelidum_tests.utils.dummy1.Dummy(attr1=1)
        dummy2 = gelidum.tests.gelidum_tests.utils.dummy2.Dummy(attr2=2)

        frozen_dummy1 = freeze(dummy1)
        frozen_dummy2 = freeze(dummy2)
        pickled_frozen_dummy1 = pickle.dumps(frozen_dummy1)
        pickled_frozen_dummy2 = pickle.dumps(frozen_dummy2)
        unpickled_frozen_dummy1 = pickle.loads(pickled_frozen_dummy1)
        unpickled_frozen_dummy2 = pickle.loads(pickled_frozen_dummy2)

        with self.assertRaises(FrozenException) as context1:
            unpickled_frozen_dummy1.attr1 = 99

        with self.assertRaises(FrozenException) as context2:
            unpickled_frozen_dummy2.attr2 = 99

        with self.assertRaises(FrozenException) as context1_get_gelidum_hot_class_module:
            unpickled_frozen_dummy2.get_gelidum_hot_class_module = 99

        with self.assertRaises(FrozenException) as context2_get_gelidum_hot_class_module:
            unpickled_frozen_dummy2.get_gelidum_hot_class_module = 99

        self.assertEqual("Can't assign attribute 'attr1' on immutable instance", str(context1.exception))
        self.assertEqual("Can't assign attribute 'attr2' on immutable instance", str(context2.exception))
        self.assertEqual(
            "Can't assign attribute 'get_gelidum_hot_class_module' on immutable instance",
            str(context1_get_gelidum_hot_class_module.exception),
        )
        self.assertEqual(
            "Can't assign attribute 'get_gelidum_hot_class_module' on immutable instance",
            str(context2_get_gelidum_hot_class_module.exception),
        )
        self.assertEqual(1, frozen_dummy1.attr1, unpickled_frozen_dummy1.attr1)
        self.assertEqual(2, frozen_dummy2.attr2, unpickled_frozen_dummy2.attr2)

        self.assertEqual("gelidum.tests.gelidum_tests.utils.dummy1", frozen_dummy1.get_gelidum_hot_class_module())
        self.assertEqual("gelidum.tests.gelidum_tests.utils.dummy2", frozen_dummy2.get_gelidum_hot_class_module())
        self.assertEqual("Dummy", frozen_dummy1.get_gelidum_hot_class_name())
        self.assertEqual("Dummy", frozen_dummy2.get_gelidum_hot_class_name())

    def test_hot_class_module_class(self):
        from gelidum.tests.gelidum_tests.utils.dummy1 import Dummy

        dummy1 = Dummy(1)
        dummy2 = Dummy(2)
        dummy3 = Dummy(3)
        frozen_dummy1 = freeze(dummy1, on_freeze="copy")
        frozen_dummy2 = freeze(dummy2, on_freeze="copy")
        frozen_dummy3 = freeze(dummy3, on_freeze="copy")

        self.assertEqual(Dummy.__name__, frozen_dummy1.get_gelidum_hot_class_name())
        self.assertEqual(Dummy.__name__, frozen_dummy2.get_gelidum_hot_class_name())
        self.assertEqual(Dummy.__name__, frozen_dummy3.get_gelidum_hot_class_name())
        self.assertEqual(Dummy.__module__, frozen_dummy1.get_gelidum_hot_class_module())
        self.assertEqual(Dummy.__module__, frozen_dummy2.get_gelidum_hot_class_module())
        self.assertEqual(Dummy.__module__, frozen_dummy3.get_gelidum_hot_class_module())

    def test_hot_class_module_internal_class(self):
        class Dummy(object):
            pass

        dummy1 = Dummy()
        dummy2 = Dummy()
        dummy3 = Dummy()
        frozen_dummy1 = freeze(dummy1, on_freeze="copy")
        frozen_dummy2 = freeze(dummy2, on_freeze="copy")
        frozen_dummy3 = freeze(dummy3, on_freeze="copy")

        self.assertEqual(Dummy.__name__, frozen_dummy1.get_gelidum_hot_class_name())
        self.assertEqual(Dummy.__name__, frozen_dummy2.get_gelidum_hot_class_name())
        self.assertEqual(Dummy.__name__, frozen_dummy3.get_gelidum_hot_class_name())
        self.assertEqual(Dummy.__module__, frozen_dummy1.get_gelidum_hot_class_module())
        self.assertEqual(Dummy.__module__, frozen_dummy2.get_gelidum_hot_class_module())
        self.assertEqual(Dummy.__module__, frozen_dummy3.get_gelidum_hot_class_module())

    def test_count_frozen_classes(self):
        class Dummy(object):
            def __init__(self, attr: int):
                self.attr = attr

        dummy1 = Dummy(1)
        dummy2 = Dummy(2)
        dummy3 = Dummy(3)
        frozen_dummy1 = freeze(dummy1, on_freeze="copy")
        frozen_dummy2 = freeze(dummy2, on_freeze="copy")
        frozen_dummy3 = freeze(dummy3, on_freeze="copy")
        frozen_classes = get_frozen_classes()

        self.assertSetEqual({frozen_dummy1.__class__}, frozen_classes)
        self.assertEqual(frozen_dummy1.__class__, frozen_dummy2.__class__, frozen_dummy3.__class__)

    def test_invalid_str_for_on_freeze_parameter(self):
        with self.assertRaises(AttributeError) as context:
            freeze(("one", 2, "three"), on_freeze="invalid")

        self.assertEqual(
            "Invalid value for on_freeze parameter, 'invalid' found, "
            "only 'copy' and 'inplace' are valid options if passed a string",
            str(context.exception),
        )

    def test_invalid_value_for_on_freeze_parameter(self):
        with self.assertRaises(AttributeError) as context:
            freeze(("one", 2, "three"), on_freeze=99)  # noqa

        self.assertEqual(
            "Invalid value for on_freeze parameter, '99' found, "
            "only 'copy', 'inplace' or a callable are valid options",
            str(context.exception),
        )

    def test_invalid_str_for_on_update_parameter(self):
        with self.assertRaises(AttributeError) as context:
            freeze(("one", 2, "three"), on_update="invalid")

        self.assertEqual(
            "Invalid value for on_update parameter, 'invalid' found, "
            "only 'exception', 'warning', and 'nothing' are valid options "
            "if passed a string",
            str(context.exception),
        )

    def test_invalid_value_for_on_update_parameter(self):
        with self.assertRaises(AttributeError) as context:
            freeze(("one", 2, "three"), on_update=1)  # noqa

        self.assertEqual(
            "Invalid value for on_update parameter, '1' found, "
            "only 'exception', 'warning', 'nothing' or a function are valid options",
            str(context.exception),
        )

    def test_structural_sharing_when_freezing_simple_objects(self):
        class Dummy(object):
            def __init__(self, value: int):
                self.value = value

        dummy = Dummy(value=1)
        frozen_dummy1 = freeze(dummy, on_update="exception", on_freeze="copy")
        frozen_dummy2 = freeze(frozen_dummy1, on_update="exception", on_freeze="copy")
        frozen_dummy3 = freeze(frozen_dummy2, on_update="exception", on_freeze="copy")

        self.assertNotEqual(id(dummy), id(frozen_dummy1))
        self.assertEqual(id(frozen_dummy1), id(frozen_dummy2))
        self.assertEqual(id(frozen_dummy2), id(frozen_dummy3))
        self.assertEqual(1, frozen_dummy1.value)
        self.assertEqual(1, frozen_dummy2.value)
        self.assertEqual(1, frozen_dummy3.value)

    def test_structural_sharing_when_freezing_same_objects_multiple_times(self):
        class DummyAttr(object):
            def __init__(self, value: str):
                self.value = value

        class Dummy(object):
            def __init__(self, dummy_attr: FrozenBase):
                self.dummy_attr: Union[FrozenBase, DummyAttr] = dummy_attr

        frozen_dummy_attr = freeze(DummyAttr("my_value"), on_update="exception", on_freeze="copy")
        dummy = Dummy(dummy_attr=frozen_dummy_attr)
        frozen_dummy1 = freeze(dummy, on_update="exception", on_freeze="copy")
        frozen_dummy2 = freeze(frozen_dummy1, on_update="exception", on_freeze="copy")
        frozen_dummy3 = freeze(frozen_dummy2, on_update="exception", on_freeze="copy")

        self.assertNotEqual(id(dummy), id(frozen_dummy1))
        self.assertEqual(id(frozen_dummy1), id(frozen_dummy2))
        self.assertEqual(id(frozen_dummy2), id(frozen_dummy3))
        self.assertEqual(id(dummy.dummy_attr), id(frozen_dummy_attr))
        self.assertEqual("my_value", dummy.dummy_attr.value)
        self.assertEqual("my_value", frozen_dummy_attr.value)
        self.assertEqual(id(frozen_dummy_attr), id(frozen_dummy1.dummy_attr))
        self.assertEqual(id(frozen_dummy1.dummy_attr), id(frozen_dummy2.dummy_attr))
        self.assertEqual(id(frozen_dummy2.dummy_attr), id(frozen_dummy3.dummy_attr))

    def test_structural_sharing_when_freezing_nested_objects(self):
        class Id(object):
            def __init__(self, value: str):
                self.value = value

        class Dummy(object):
            def __init__(self, value: int, obj_id: Id):
                self.value = value
                self.obj_id = freeze(obj_id, on_update="exception", on_freeze="copy")

        dummy = Dummy(value=1, obj_id=Id("unique_id"))
        frozen_dummy = freeze(dummy, on_update="exception", on_freeze="copy")

        self.assertNotEqual(id(dummy), id(frozen_dummy))
        self.assertEqual(id(dummy.obj_id.value), id(frozen_dummy.obj_id.value))
        self.assertEqual(id(dummy.obj_id), id(frozen_dummy.obj_id))
        self.assertEqual("unique_id", dummy.obj_id.value)
        self.assertEqual("unique_id", frozen_dummy.obj_id.value)

    def test_structural_sharing_immutable_list(self):
        class Dummy(object):
            def __init__(self, value: int):
                self.value = value

        class ConsList(object):
            @staticmethod
            def __freeze(item: Any) -> FrozenBase:
                return freeze(item, on_update="exception", on_freeze="copy")

            def __init__(self, *args):
                if len(args) > 0:
                    self._items = tuple(self.__freeze(arg_i) for arg_i in args)
                else:
                    self._items = tuple()

            def __getitem__(self, key) -> Any:
                return self._items[key]

            def __len__(self) -> int:
                return len(self._items)

            def __add__(self, other) -> "ConsList":
                frozen_other = freeze(other, on_update="exception", on_freeze="copy")
                return ConsList(*(self._items + (frozen_other,)))

        immutable_list_size_0 = ConsList()
        immutable_list_size_1 = immutable_list_size_0 + 1
        immutable_list_size_2 = immutable_list_size_1 + 2
        immutable_list_size_3 = immutable_list_size_2 + 3
        immutable_list_size_4 = immutable_list_size_3 + Dummy(5)
        immutable_list_size_6 = immutable_list_size_4 + Dummy(5) + "a string"

        self.assertNotEqual(id(immutable_list_size_1), id(immutable_list_size_2))
        self.assertNotEqual(id(immutable_list_size_2), id(immutable_list_size_3))
        self.assertNotEqual(id(immutable_list_size_3), id(immutable_list_size_4))
        self.assertNotEqual(id(immutable_list_size_4), id(immutable_list_size_6))
        self.assertEqual(1, len(immutable_list_size_1))
        self.assertEqual(2, len(immutable_list_size_2))
        self.assertEqual(3, len(immutable_list_size_3))
        self.assertEqual(4, len(immutable_list_size_4))
        self.assertEqual(6, len(immutable_list_size_6))
        self.assertEqual(id(immutable_list_size_1[0]), id(immutable_list_size_2[0]))
        self.assertEqual(id(immutable_list_size_1[0]), id(immutable_list_size_3[0]))
        self.assertEqual(id(immutable_list_size_1[0]), id(immutable_list_size_4[0]))
        self.assertEqual(id(immutable_list_size_1[0]), id(immutable_list_size_6[0]))
        self.assertEqual(id(immutable_list_size_2[1]), id(immutable_list_size_3[1]))
        self.assertEqual(id(immutable_list_size_2[1]), id(immutable_list_size_4[1]))
        self.assertEqual(id(immutable_list_size_2[1]), id(immutable_list_size_6[1]))
        self.assertEqual(id(immutable_list_size_3[2]), id(immutable_list_size_4[2]))
        self.assertEqual(id(immutable_list_size_3[2]), id(immutable_list_size_6[2]))
        self.assertEqual(id(immutable_list_size_4[3]), id(immutable_list_size_6[3]))

    def test_use_frozen_object_as_dict_key(self):
        class Dummy(object):
            def __init__(self, value: int):
                self.attr = value

        dummy1 = Dummy(value=1)
        frozen_dummy1 = freeze(dummy1, on_freeze="copy")
        dummy2 = Dummy(value=2)
        frozen_dummy2 = freeze(dummy2, on_freeze="copy")
        my_dict = {frozen_dummy1: dummy1, frozen_dummy2: dummy2}

        self.assertEqual(2, len(my_dict))
        self.assertEqual({frozen_dummy1, frozen_dummy2}, set(my_dict.keys()))
        self.assertTrue(frozen_dummy1 in my_dict)
        self.assertEqual(dummy1, my_dict[frozen_dummy1])
        self.assertTrue(frozen_dummy2 in my_dict)
        self.assertEqual(dummy2, my_dict[frozen_dummy2])

    @unittest.skipUnless(NUMPY_INSTALLED, "numpy is not installed, TestFreeze.test_freeze_ndarray test skipped")
    def test_freeze_ndarray(self):
        import numpy as np

        array = np.array([1, 2, 3])
        frozen_array = freeze(array, on_freeze="copy")
        frozen_array_copy = frozen_array.copy()

        self.assertEqual(array.shape, frozen_array.shape)
        self.assertEqual(array.shape, frozen_array_copy.shape)
        self.assertEqual(array[0], frozen_array[0])
        self.assertEqual(array[1], frozen_array[1])
        self.assertEqual(array[2], frozen_array[2])
        self.assertEqual(array[0], frozen_array_copy[0])
        self.assertEqual(array[1], frozen_array_copy[1])
        self.assertEqual(array[2], frozen_array_copy[2])
        self.assertTrue(isinstance(frozen_array, FrozenBase))
        self.assertTrue(isinstance(frozen_array, np.ndarray))
        self.assertTrue(isinstance(frozen_array_copy, FrozenBase))
        self.assertTrue(isinstance(frozen_array_copy, np.ndarray))



import logging
import unittest
from unittest import mock
from typing import List, Any
from unittest.mock import call

from gelidum import OnFreezeCopier, OnFreezeOriginalObjTracker, OnFreezeIdentityFunc
from gelidum import freeze
from gelidum.frozen import FrozenBase, clear_frozen_classes


class TestOnFreeze(unittest.TestCase):
    def setUp(self) -> None:
        clear_frozen_classes()

    def test_on_freeze_original_obj_tracking(self):
        class DummyAttr1(object):
            def __init__(self, value: int):
                self.attr = value

        class DummyAttr2(object):
            def __init__(self, value: int):
                self.attr = value

        class Dummy(object):
            def __init__(self, value1: int, value2: int):
                self.attr1 = DummyAttr1(value1)
                self.attr2 = DummyAttr2(value2)

        dummy1 = Dummy(value1=1, value2=2)

        freezer = OnFreezeOriginalObjTracker()

        frozen_dummy1 = freeze(dummy1, on_freeze=freezer)
        original_obj = freezer.original_obj

        self.assertEqual(id(dummy1), id(original_obj))
        self.assertNotEqual(id(dummy1), id(frozen_dummy1))
        self.assertTrue(isinstance(original_obj, Dummy))
        self.assertTrue(isinstance(original_obj.attr1, DummyAttr1))
        self.assertTrue(isinstance(original_obj.attr2, DummyAttr2))
        self.assertIs(original_obj.__class__, Dummy)
        self.assertIs(original_obj.attr1.__class__, DummyAttr1)
        self.assertIs(original_obj.attr2.__class__, DummyAttr2)
        self.assertEqual(1, original_obj.attr1.attr)
        self.assertEqual(2, original_obj.attr2.attr)
        self.assertTrue(isinstance(frozen_dummy1, Dummy))
        self.assertTrue(isinstance(frozen_dummy1.attr1, DummyAttr1))
        self.assertTrue(isinstance(frozen_dummy1.attr2, DummyAttr2))
        self.assertTrue(isinstance(frozen_dummy1, FrozenBase))
        self.assertTrue(isinstance(frozen_dummy1.attr1, FrozenBase))
        self.assertTrue(isinstance(frozen_dummy1.attr2, FrozenBase))
        self.assertIsNot(frozen_dummy1.__class__, Dummy)
        self.assertIsNot(frozen_dummy1.attr1.__class__, DummyAttr1)
        self.assertIsNot(frozen_dummy1.attr2.__class__, DummyAttr2)
        self.assertEqual(1, frozen_dummy1.attr1.attr)
        self.assertEqual(2, frozen_dummy1.attr2.attr)

    def test_on_freeze_logging(self):
        class DummyAttr1(object):
            def __init__(self, value: int):
                self.attr = value

        class DummyAttr2(object):
            def __init__(self, value: int):
                self.attr = value

        class Dummy(object):
            def __init__(self, value1: int, value2: int):
                self.attr1 = DummyAttr1(value1)
                self.attr2 = DummyAttr2(value2)

        dummy1 = Dummy(value1=1, value2=2)

        class OnFreezeLogger(OnFreezeCopier):
            def __init__(self, log: logging.Logger):
                self.__log = log

            def __call__(self, obj: Any) -> Any:
                self.__log.debug(f"{obj.__class__.__name__} object frozen")
                return super().__call__(obj=obj)

        mock_log = mock.Mock()
        freezer = OnFreezeLogger(log=mock_log)

        frozen_dummy1 = freeze(dummy1, on_freeze=freezer)

        self.assertEqual(
            [call("Dummy object frozen"), call("DummyAttr1 object frozen"), call("DummyAttr2 object frozen")],
            mock_log.debug.call_args_list,
        )
        self.assertTrue(isinstance(frozen_dummy1, Dummy))
        self.assertTrue(isinstance(frozen_dummy1.attr1, DummyAttr1))
        self.assertTrue(isinstance(frozen_dummy1.attr2, DummyAttr2))
        self.assertTrue(isinstance(frozen_dummy1, FrozenBase))
        self.assertTrue(isinstance(frozen_dummy1.attr1, FrozenBase))
        self.assertTrue(isinstance(frozen_dummy1.attr2, FrozenBase))
        self.assertIsNot(frozen_dummy1.__class__, Dummy)
        self.assertIsNot(frozen_dummy1.attr1.__class__, DummyAttr1)
        self.assertIsNot(frozen_dummy1.attr2.__class__, DummyAttr2)
        self.assertEqual(1, frozen_dummy1.attr1.attr)
        self.assertEqual(2, frozen_dummy1.attr2.attr)

    def test_on_freeze_custom_freezer_in_place(self):
        class DummyAttr1(object):
            def __init__(self, value: int):
                self.attr = value

        class DummyAttr2(object):
            def __init__(self, value: int):
                self.attr = value

        class Dummy(object):
            def __init__(self, value1: int, value2: int):
                self.attr1 = DummyAttr1(value1)
                self.attr2 = DummyAttr2(value2)

        dummy1 = Dummy(value1=1, value2=2)

        class OnFreezeFullTrackingInPlace(OnFreezeIdentityFunc):
            def __init__(self):
                self.__objs: List[Any] = []

            def __call__(self, obj: Any) -> Any:
                self.__objs.append(obj)
                return super().__call__(obj=obj)

            @property
            def original_objs(self) -> List[Any]:
                return self.__objs

        freezer = OnFreezeFullTrackingInPlace()

        frozen_dummy1 = freeze(dummy1, on_freeze=freezer)
        original_objs = freezer.original_objs

        self.assertEqual(3, len(original_objs))
        self.assertEqual(id(dummy1), id(original_objs[0]))
        self.assertEqual(id(dummy1.attr1), id(original_objs[1]))
        self.assertEqual(id(dummy1.attr2), id(original_objs[2]))
        self.assertEqual(id(original_objs[0].attr1), id(original_objs[1]))
        self.assertEqual(id(original_objs[0].attr2), id(original_objs[2]))
        self.assertEqual(id(dummy1), id(frozen_dummy1))
        self.assertTrue(isinstance(original_objs[0], Dummy))
        self.assertTrue(isinstance(original_objs[1], DummyAttr1))
        self.assertTrue(isinstance(original_objs[2], DummyAttr2))
        self.assertEqual(1, original_objs[1].attr)
        self.assertEqual(2, original_objs[2].attr)
        self.assertTrue(isinstance(frozen_dummy1, Dummy))
        self.assertTrue(isinstance(frozen_dummy1.attr1, DummyAttr1))
        self.assertTrue(isinstance(frozen_dummy1.attr2, DummyAttr2))
        self.assertTrue(isinstance(frozen_dummy1, FrozenBase))
        self.assertTrue(isinstance(frozen_dummy1.attr1, FrozenBase))
        self.assertTrue(isinstance(frozen_dummy1.attr2, FrozenBase))
        self.assertIsNot(frozen_dummy1.__class__, Dummy)
        self.assertIsNot(frozen_dummy1.attr1.__class__, DummyAttr1)
        self.assertIsNot(frozen_dummy1.attr2.__class__, DummyAttr2)
        self.assertEqual(1, frozen_dummy1.attr1.attr)
        self.assertEqual(2, frozen_dummy1.attr2.attr)



from typing import Any, Callable, Optional, Union, Sequence, Generator, Iterable

from gelidum.exceptions import FrozenException
from gelidum.frozen import FrozenBase
from gelidum.typing import FrozenType, FrozenZet

__all__ = ["frozenzet"]

_FrozenZetParameterType = Optional[Union[Sequence, Generator, Iterable]]


class frozenzet(frozenset, FrozenBase):  # noqa
    def __raise_immutable_exception(self, *args, **kwargs):
        raise FrozenException("'frozenzet' object is immutable")

    def __new__(
        cls, seq: Optional[_FrozenZetParameterType] = None, freeze_func: Optional[Callable[[Any], FrozenBase]] = None
    ) -> "frozenzet":
        if freeze_func is None:

            def freeze_func(item: Any) -> FrozenType:
                from gelidum.freeze import freeze

                return freeze(item, on_update="exception", on_freeze="copy")

        if seq:
            self = frozenset.__new__(cls, (freeze_func(arg) for arg in seq))
        else:
            self = frozenset.__new__(cls, [])
        return self

    def __init__(
        self, seq: Optional[_FrozenZetParameterType] = None, freeze_func: Optional[Callable[[Any], FrozenBase]] = None
    ):
        super().__init__()

    @classmethod
    def _gelidum_on_update(cls, *args, **kwargs):
        raise FrozenException("'frozenzet' object is immutable")

    @classmethod
    def get_gelidum_hot_class_name(cls) -> str:
        return "frozenset"

    @classmethod
    def get_gelidum_hot_class_module(cls) -> str:
        return "builtins.frozenset"

    def __hash__(self) -> int:
        return hash(tuple(v for v in self))

    def __add__(self, other: FrozenZet) -> FrozenZet:
        joined_set = set()
        for item in self:
            joined_set.add(item)
        for item in other:
            joined_set.add(item)
        return frozenzet(joined_set)

    def add(self, item) -> None:
        self.__raise_immutable_exception()

    def remove(self, item) -> None:
        self.__raise_immutable_exception()

    def discard(self, item) -> None:
        self.__raise_immutable_exception()

    def pop(self) -> None:
        self.__raise_immutable_exception()

    def clear(self) -> None:
        self.__raise_immutable_exception()

    def update(self, *others) -> None:
        self.__raise_immutable_exception()

    def __ior__(self, *others) -> None:
        self.__raise_immutable_exception()

    def intersection_update(self, *others) -> None:
        self.__raise_immutable_exception()

    def __iand__(self, *others) -> None:
        self.__raise_immutable_exception()

    def difference_update(self, *others) -> None:
        self.__raise_immutable_exception()

    def __isub__(self, *others) -> None:
        self.__raise_immutable_exception()

    def symmetric_difference_update(self, others) -> None:
        self.__raise_immutable_exception()

    def __ixor__(self, *others) -> None:
        self.__raise_immutable_exception()

    def copy(self) -> "frozenzet":
        """
        frozenzet objects are only shallow-copied.
        """
        return self



from __future__ import annotations

import functools
from typing import Optional, Iterable, Set, Any
from gelidum import freeze


def freeze_params(params: Optional[Iterable[str]] = None):
    def inner_freeze_params(func):
        """Freeze all input params of a method"""

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            func_args = tuple([freeze(arg, on_freeze="copy") for arg in args])
            func_kwargs = {
                kwarg_name: freeze(kwarg, on_freeze="copy") if kwarg_name in params else kwarg
                for kwarg_name, kwarg in kwargs.items()
            }
            return func(*func_args, **func_kwargs)

        return wrapper

    return inner_freeze_params


def freeze_final(func):
    """Freeze all final params of a method"""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        unnamed_params_to_freeze: Set[int] = {
            i
            for i, (param_name, param_typing) in enumerate(func.__annotations__.items())
            if __param_is_final(param_typing)
        }
        named_params_to_freeze: Set[str] = {
            param_name for param_name, param_typing in func.__annotations__.items() if __param_is_final(param_typing)
        }
        func_args = tuple(
            [
                (freeze(arg, on_freeze="copy") if arg_index in unnamed_params_to_freeze else arg)
                for arg_index, arg in enumerate(args)
            ]
        )
        func_kwargs = {
            kwarg_name: (freeze(kwarg, on_freeze="copy") if kwarg_name in named_params_to_freeze else kwarg)
            for kwarg_name, kwarg in kwargs.items()
        }
        return func(*func_args, **func_kwargs)

    return wrapper


def __param_is_final(param_typing: Any) -> bool:
    param_typing_str = str(param_typing)
    return param_typing_str.startswith("typing.Final") or param_typing_str.startswith("gelidum.typing.Final")



import unittest
from typing import Any

from gelidum import FrozenException, freeze
from gelidum.collections.frozenzet import frozenzet
from gelidum.frozen import FrozenBase


class TestFrozenzet(unittest.TestCase):  # noqa
    def test_empty_construction(self):
        frozen_zet = frozenzet()

        self.assertEqual(0, len(frozen_zet))

    def test_construction(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_zet = frozenzet([Dummy(1), Dummy(2), Dummy(3)])
        unfrozen_list = sorted(list(frozen_zet), key=lambda d: d.value)

        self.assertTrue(isinstance(frozen_zet, FrozenBase))
        self.assertEqual(3, len(frozen_zet))
        self.assertEqual(3, len(unfrozen_list))
        self.assertTrue(isinstance(unfrozen_list[0], FrozenBase))
        self.assertEqual(1, unfrozen_list[0].value)
        self.assertTrue(isinstance(unfrozen_list[1], FrozenBase))
        self.assertEqual(2, unfrozen_list[1].value)
        self.assertTrue(isinstance(unfrozen_list[2], FrozenBase))
        self.assertEqual(3, unfrozen_list[2].value)

    def test_construction_from_tuple(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        dummy = Dummy(3)
        frozen_dummy = freeze(dummy)
        frozen_zet = frozenzet((1, "2", frozen_dummy))

        self.assertTrue(isinstance(frozen_zet, FrozenBase))
        self.assertEqual(3, len(frozen_zet))
        self.assertTrue(1 in frozen_zet)
        self.assertTrue("2" in frozen_zet)
        self.assertFalse(dummy in frozen_zet)
        self.assertTrue(frozen_dummy in frozen_zet)

    def test_construction_from_generator(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        def generator():
            yield Dummy(1)
            yield Dummy(2)
            yield Dummy(3)

        frozen_zet = frozenzet(generator())
        unfrozen_list = sorted(list(frozen_zet), key=lambda d: d.value)

        self.assertTrue(isinstance(frozen_zet, FrozenBase))
        self.assertEqual(3, len(frozen_zet))
        self.assertEqual(3, len(unfrozen_list))
        self.assertTrue(isinstance(unfrozen_list[0], FrozenBase))
        self.assertEqual(1, unfrozen_list[0].value)
        self.assertTrue(isinstance(unfrozen_list[1], FrozenBase))
        self.assertEqual(2, unfrozen_list[1].value)
        self.assertTrue(isinstance(unfrozen_list[2], FrozenBase))
        self.assertEqual(3, unfrozen_list[2].value)

    def test_castings(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

            def __str__(self) -> str:
                return f"{self.value}"

        dummy = Dummy(3)
        frozen_dummy = freeze(dummy)

        self.assertEqual({1, "2", frozen_dummy}, set(frozenzet((1, "2", frozen_dummy))))
        self.assertEqual([1, "2", frozen_dummy], sorted(list(frozenzet((1, "2", frozen_dummy))), key=lambda e: str(e)))

    def test_setattr_exception(self):
        frozen_zet = frozenzet([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            setattr(frozen_zet, "my_attr", "value")

        self.assertEqual("'frozenzet' object is immutable", str(context.exception))

    def test_len(self):
        frozen_zet = frozenzet([1, 2, 3])

        self.assertEqual(3, len(frozen_zet))

    def test_add(self):
        frozen_zet = frozenzet([1, 2, 3])
        unfrozen_list = list(frozen_zet)
        frozen_zet2 = frozen_zet + frozenzet([1, 2]) + frozenzet([2, 3, 4])
        unfrozen_list2 = list(frozen_zet2)
        unfrozen_set2 = set(frozen_zet2)

        self.assertTrue(isinstance(frozen_zet, FrozenBase))
        self.assertEqual(3, len(frozen_zet))
        self.assertEqual(3, len(unfrozen_list))
        self.assertEqual(1, unfrozen_list[0])
        self.assertEqual(2, unfrozen_list[1])
        self.assertEqual(3, unfrozen_list[2])
        self.assertTrue(1 in frozen_zet)
        self.assertTrue(2 in frozen_zet)
        self.assertTrue(3 in frozen_zet)
        self.assertTrue(isinstance(frozen_zet2, FrozenBase))
        self.assertEqual(4, len(frozen_zet2))
        self.assertEqual(4, len(unfrozen_list2))
        self.assertEqual(4, len(unfrozen_set2))
        self.assertEqual(1, unfrozen_list2[0])
        self.assertEqual(2, unfrozen_list2[1])
        self.assertEqual(3, unfrozen_list2[2])
        self.assertEqual(4, unfrozen_list2[3])
        self.assertTrue(1 in unfrozen_set2)
        self.assertTrue(2 in unfrozen_set2)
        self.assertTrue(3 in unfrozen_set2)
        self.assertTrue(4 in unfrozen_set2)
        self.assertTrue(1 in frozen_zet2)
        self.assertTrue(2 in frozen_zet2)
        self.assertTrue(3 in frozen_zet2)
        self.assertTrue(4 in frozen_zet2)

    def test_hash(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_zet = frozenzet([1, 2, 3, [4, 5, 6], Dummy(2), frozenzet({6, 7})])

        self.assertTrue(isinstance(hash(frozen_zet), int))

    def test_contains(self):
        frozen_zet = frozenzet([1, 2, 3])

        self.assertTrue(1 in frozen_zet)
        self.assertFalse(1 not in frozen_zet)
        self.assertTrue(9 not in frozen_zet)
        self.assertFalse(9 in frozen_zet)

    def test_subset_superset(self):
        frozen_zet = frozenzet([1, 2, 3])
        frozen_zet_copy = frozenzet([1, 2, 3])
        superset_frozen_zet = frozenzet([1, 2, 3, 4, 5, 6])

        self.assertTrue(superset_frozen_zet.issuperset(frozen_zet))
        self.assertTrue(superset_frozen_zet >= frozen_zet)
        self.assertFalse(frozen_zet_copy > frozen_zet)
        self.assertTrue(frozen_zet_copy >= frozen_zet)
        self.assertTrue(superset_frozen_zet > frozen_zet)
        self.assertTrue(frozen_zet.issubset(superset_frozen_zet))
        self.assertTrue(frozen_zet <= superset_frozen_zet)
        self.assertTrue(frozen_zet < superset_frozen_zet)
        self.assertFalse(frozen_zet_copy < frozen_zet)
        self.assertTrue(frozen_zet_copy <= frozen_zet)

    def test_min(self):
        frozen_zet = frozenzet([1, 2, 3])

        self.assertEqual(1, min(frozen_zet))

    def test_max(self):
        frozen_zet = frozenzet([1, 2, 3])

        self.assertEqual(3, max(frozen_zet))

    def test_iter(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dummy = freeze(Dummy(9))
        frozen_zet = frozenzet([1, 2, 3, frozen_dummy])
        items = set()
        for item in frozen_zet:
            items.add(item)

        self.assertSetEqual({1, 2, 3, frozen_dummy}, items)

    def test_next(self):
        class Dummy:
            def __init__(self, value: Any):
                self.value = value

        frozen_dummy = freeze(Dummy(9))
        frozen_zet = frozenzet([1, 2, 3, frozen_dummy])
        frozen_list_iter = iter(frozen_zet)

        items = set()
        for _ in frozen_zet:
            items.add(next(frozen_list_iter))

        self.assertSetEqual({1, 2, 3, frozen_dummy}, items)

    def test_copy(self):
        frozen_zet = frozenzet([1, 2, 3])

        frozen_zet_copy = frozen_zet.copy()

        self.assertEqual(id(frozen_zet_copy), id(frozen_zet))

    def test_add_exception(self):
        frozen_zet = frozenzet([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_zet.add(11)

        self.assertEqual("'frozenzet' object is immutable", str(context.exception))

    def test_remove_exception(self):
        frozen_zet = frozenzet([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_zet.remove(2)

        self.assertEqual("'frozenzet' object is immutable", str(context.exception))

    def test_pop_exception(self):
        frozen_zet = frozenzet([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_zet.pop()

        self.assertEqual("'frozenzet' object is immutable", str(context.exception))

    def test_clear_exception(self):
        frozen_zet = frozenzet([1, 2, 3])

        with self.assertRaises(FrozenException) as context:
            frozen_zet.clear()

        self.assertEqual("'frozenzet' object is immutable", str(context.exception))

    def test_update_exception(self):
        frozen_zet = frozenzet([1, 2, 3])
        frozen_zet2 = frozenzet([4, 5, 6])

        with self.assertRaises(FrozenException) as context1:
            frozen_zet.update(frozen_zet2)

        with self.assertRaises(FrozenException) as context2:
            frozen_zet |= frozen_zet2

        self.assertEqual("'frozenzet' object is immutable", str(context1.exception))
        self.assertEqual("'frozenzet' object is immutable", str(context2.exception))

    def test_intersection_update_exception(self):
        frozen_zet = frozenzet([1, 2, 3])
        frozen_zet2 = frozenzet([4, 5, 6])

        with self.assertRaises(FrozenException) as context1:
            frozen_zet.intersection_update(frozen_zet2)

        with self.assertRaises(FrozenException) as context2:
            frozen_zet &= frozen_zet2

        self.assertEqual("'frozenzet' object is immutable", str(context1.exception))
        self.assertEqual("'frozenzet' object is immutable", str(context2.exception))

    def test_difference_update_exception(self):
        frozen_zet = frozenzet([1, 2, 3])
        frozen_zet2 = frozenzet([4, 5, 6])

        with self.assertRaises(FrozenException) as context1:
            frozen_zet.difference_update(frozen_zet2)

        with self.assertRaises(FrozenException) as context2:
            frozen_zet -= frozen_zet2

        self.assertEqual("'frozenzet' object is immutable", str(context1.exception))
        self.assertEqual("'frozenzet' object is immutable", str(context2.exception))

    def test_symmetric_difference_update_exception(self):
        frozen_zet = frozenzet([1, 2, 3])
        frozen_zet2 = frozenzet([4, 5, 6])

        with self.assertRaises(FrozenException) as context1:
            frozen_zet.symmetric_difference_update(frozen_zet2)

        with self.assertRaises(FrozenException) as context2:
            frozen_zet ^= frozen_zet2

        self.assertEqual("'frozenzet' object is immutable", str(context1.exception))
        self.assertEqual("'frozenzet' object is immutable", str(context2.exception))



from typing import Any, Callable, Optional

import numpy as np

from gelidum.exceptions import FrozenException
from gelidum.frozen import FrozenBase


__all__ = ["frozenndarray"]


class frozenndarray(np.ndarray, FrozenBase):  # noqa
    """
    Read https://numpy.org/devdocs/user/basics.subclassing.html for more
    information about numpy.ndarray subclassing.
    """

    def __new__(cls, ndarray: np.ndarray, freeze_func: Optional[Callable[[Any], FrozenBase]] = None, *args, **kwargs):
        obj = ndarray.copy().view(cls)
        obj.flags.writeable = False
        return obj

    @classmethod
    def _gelidum_on_update(cls, *args, **kwargs):
        raise FrozenException("'ndarray' object is immutable")

    @classmethod
    def get_gelidum_hot_class_name(cls) -> str:
        return "ndarray"

    @classmethod
    def get_gelidum_hot_class_module(cls) -> str:
        return "numpy.ndarray"



import copy
from typing import Any, Optional, Union

from gelidum.typing import OnFreezeFuncType


class OnFreezeCopier:
    def __call__(self, obj: Any) -> Any:
        return copy.deepcopy(obj)


class OnFreezeIdentityFunc:
    def __call__(self, obj: Any) -> Any:
        return obj


class OnFreezeOriginalObjTracker(OnFreezeCopier):
    """
    A callable class that stores the first object that was
    frozen. Useful for keeping track of original "hot" objects.
    """

    def __init__(self):
        self.__original_obj: Optional[Any] = None

    def __call__(self, obj: Any) -> Any:
        if self.__original_obj is None:
            self.__original_obj = obj
        return super().__call__(obj=obj)

    @property
    def original_obj(self) -> Optional[Any]:
        return self.__original_obj


_ON_FREEZE_COPIER = OnFreezeCopier()
_ON_FREEZE_IDENTITY_FUNC = OnFreezeIdentityFunc()


def on_freeze_func_creator(on_freeze: Union[str, OnFreezeFuncType]) -> OnFreezeFuncType:
    if isinstance(on_freeze, str):
        if on_freeze == "copy":
            return _ON_FREEZE_COPIER
        elif on_freeze == "inplace":
            return _ON_FREEZE_IDENTITY_FUNC
        else:
            raise AttributeError(
                f"Invalid value for on_freeze parameter, '{on_freeze}' found, "
                f"only 'copy' and 'inplace' are valid options if passed a string"
            )

    elif callable(on_freeze):
        return on_freeze

    else:
        raise AttributeError(
            f"Invalid value for on_freeze parameter, '{on_freeze}' found, "
            f"only 'copy', 'inplace' or a callable are valid options"
        )



import sys
import threading
import uuid
from typing import cast, Type, Dict, Optional, Set, Iterable, Any
from gelidum.typing import OnUpdateFuncType
from gelidum.frozen.frozen_base import FrozenBase


def __create_frozen_class(
    klass: Type[object], attrs: Iterable[str], on_update_func: OnUpdateFuncType
) -> Type[FrozenBase]:
    camel_case_module = klass.__module__.title().replace(".", "").replace("_", "")
    frozen_class_name = f"Frozen{klass.__name__}From{camel_case_module}"
    frozen_class: Type[FrozenBase] = cast(
        Type[FrozenBase],
        type(
            frozen_class_name,
            (FrozenBase, klass),
            {
                **{
                    "get_gelidum_hot_class_name": lambda _: klass.__name__,
                    "get_gelidum_hot_class_module": lambda _: klass.__module__,
                    "_gelidum_on_update": lambda _self, *args, **kwargs: on_update_func(*args, **kwargs),
                    "original_obj": None,
                    **{attr: None for attr in attrs},
                }
            },
        ),
    )
    __store_frozen_class(klass=klass, frozen_class=frozen_class)
    return frozen_class


def make_frozen_class(klass: Type[object], attrs: Iterable[str], on_update: OnUpdateFuncType) -> Type[FrozenBase]:
    frozen_class = get_frozen_class(klass_key=f"{klass.__module__}.{klass.__qualname__}")

    if not frozen_class:
        frozen_class = __create_frozen_class(klass=klass, attrs=attrs, on_update_func=on_update)

    return frozen_class


def make_unique_class(klass: Type[object], attrs: Dict[str, Any], on_update: OnUpdateFuncType) -> Type[FrozenBase]:

    camel_case_module = klass.__module__.title().replace(".", "").replace("_", "")
    unique_suffix = str(uuid.uuid4()).replace("-", "")
    frozen_class_name = f"Frozen{klass.__name__}From{camel_case_module}{unique_suffix}"
    frozen_class: Type[FrozenBase] = cast(
        Type[FrozenBase],
        type(
            frozen_class_name,
            (FrozenBase, klass),
            {
                **{
                    "get_gelidum_hot_class_name": lambda _: klass.__name__,
                    "get_gelidum_hot_class_module": lambda _: klass.__module__,
                    "_gelidum_on_update": lambda _self, *args, **kwargs: on_update(*args, **kwargs),
                    **attrs,
                    "__init__": lambda _: None,
                }
            },
        ),
    )

    return frozen_class


__FROZEN_CLASSES: Dict[str, Type[FrozenBase]] = dict()
__FROZEN_CLASSES_LOCK = threading.Lock()


def get_frozen_class(klass_key: str) -> Optional[Type[FrozenBase]]:
    with __FROZEN_CLASSES_LOCK:
        frozen_class = __FROZEN_CLASSES.get(klass_key)
    return frozen_class


def get_frozen_classes() -> Set[Type[FrozenBase]]:
    with __FROZEN_CLASSES_LOCK:
        return set(__FROZEN_CLASSES.values())


def __store_frozen_class(klass: Type[object], frozen_class: Type[FrozenBase]) -> None:
    """
    Add a frozen class to this module.
    Required for pickle serialization as only objects of non-dynamic
    classes are allowed.
    This method is thread-safe.
    :param klass: the original class.
    :param frozen_class: a class that inherits from FrozenBase.
    """
    with __FROZEN_CLASSES_LOCK:
        klass_key = f"{klass.__module__}.{klass.__qualname__}"
        __FROZEN_CLASSES[klass_key] = frozen_class
        # Required for pickling frozen objects (only classes defined in actual
        # modules can have their objects pickled)
        setattr(sys.modules[__name__], frozen_class.__name__, frozen_class)


def clear_frozen_classes() -> None:
    with __FROZEN_CLASSES_LOCK:
        __FROZEN_CLASSES.clear()



import importlib.util
from importlib.machinery import ModuleSpec


def package_is_installed(package_name: str) -> bool:
    return isinstance(importlib.util.find_spec(package_name), ModuleSpec)


NUMPY_INSTALLED = package_is_installed("numpy")



import time
import unittest
from gelidum import freeze
from gelidum.frozen import clear_frozen_classes


class TestTimePerformance(unittest.TestCase):
    def setUp(self) -> None:
        clear_frozen_classes()

    def test_freeze_inplace_or_not(self):
        class DummyAttrLevel3(object):
            def __init__(self, attr: str):
                self.attr = attr

        class DummyAttrLevel2(object):
            def __init__(self, attr: str):
                self.attr = DummyAttrLevel3(attr)

        class DummyAttrLevel1(object):
            def __init__(self, attr: str):
                self.attr = DummyAttrLevel2(attr)

        class Dummy(object):
            def __init__(self):
                for attr_index in range(0, 100_000):
                    setattr(self, f"attr{attr_index+1}", DummyAttrLevel1(str(attr_index) * 1_000))

        dummy1 = Dummy()
        start_inplace = time.time()
        freeze(dummy1, on_freeze="inplace")
        spent_time_inplace = time.time() - start_inplace
        dummy2 = Dummy()
        start_not_inplace = time.time()
        freeze(dummy2, on_freeze="copy")
        spent_time_not_inplace = time.time() - start_not_inplace

        self.assertLessEqual(spent_time_inplace, spent_time_not_inplace)

    def test_many_attr_class_freeze(self):
        class BigDummy(object):
            def __init__(self, attr: str):
                for attr_index in range(0, 10000):
                    setattr(self, f"attr{attr_index+1}", attr)

        value = "1" * 1_000_000
        dummy = BigDummy(attr=value)

        spent_times = []
        for i in range(5):
            start = time.time()
            freeze(dummy, on_freeze="copy")
            spent_times.append(time.time() - start)

        spent_time = sum(spent_times) / len(spent_times)

        self.assertLessEqual(spent_time, 0.4)



import unittest

from gelidum import NUMPY_INSTALLED, freeze
from gelidum.frozen import FrozenBase


@unittest.skipUnless(NUMPY_INSTALLED, "numpy is not installed, TestFrozenndarray tests skipped")
class TestFrozenndarray(unittest.TestCase):  # noqa
    def test_freeze_int64_ndarray(self):
        import numpy as np

        array = np.array([1, 2, 3], dtype=np.int64).reshape((1, 3))
        frozen_array = freeze(array, on_freeze="copy")
        frozen_array_copy = frozen_array.copy()

        self.assertEqual(array.shape, frozen_array.shape)
        self.assertEqual(array.shape, frozen_array_copy.shape)
        self.assertEqual(array.tolist(), frozen_array.tolist())
        self.assertEqual(array.tolist(), frozen_array_copy.tolist())
        self.assertTrue(isinstance(frozen_array, FrozenBase))
        self.assertTrue(isinstance(frozen_array, np.ndarray))
        self.assertTrue(isinstance(frozen_array_copy, FrozenBase))
        self.assertTrue(isinstance(frozen_array_copy, np.ndarray))

    def test_freeze_int64_custom_shape_ndarray(self):
        import numpy as np

        array = np.array([1, 2, 3, 4], dtype=np.int64).reshape((2, 2))
        frozen_array = freeze(array, on_freeze="copy")
        frozen_array_copy = frozen_array.copy()

        self.assertEqual(array.shape, frozen_array.shape)
        self.assertEqual(array.shape, frozen_array_copy.shape)
        self.assertEqual(array.tolist(), frozen_array.tolist())
        self.assertEqual(array.tolist(), frozen_array_copy.tolist())
        self.assertTrue(isinstance(frozen_array, FrozenBase))
        self.assertTrue(isinstance(frozen_array, np.ndarray))
        self.assertTrue(isinstance(frozen_array_copy, FrozenBase))
        self.assertTrue(isinstance(frozen_array_copy, np.ndarray))

    def test_assignment_single_value(self):
        import numpy as np

        array = np.array([1, 2, 3, 4], dtype=np.int64).reshape((2, 2))
        frozen_array = freeze(array, on_freeze="copy")

        with self.assertRaises(ValueError) as context:
            frozen_array[0, 1] = 99

        array[0, 1] = 99

        self.assertEqual("assignment destination is read-only", str(context.exception))
        self.assertEqual(2, frozen_array[0, 1])
        self.assertEqual(99, array[0, 1])

    def test_assignment_multiple_values(self):
        import numpy as np

        array = np.array([1, 2, 3, 4], dtype=int).reshape((2, 2))
        frozen_array = freeze(array, on_freeze="copy")

        with self.assertRaises(ValueError) as context:
            frozen_array[:, 0] = 99

        array[:, 0] = 99

        self.assertEqual("assignment destination is read-only", str(context.exception))
        self.assertEqual(99, array[0, 0])
        self.assertEqual(99, array[1, 0])
        self.assertEqual(1, frozen_array[0, 0])
        self.assertEqual(3, frozen_array[1, 0])



from gelidum.dependencies import NUMPY_INSTALLED
from gelidum.collections.frozendict import frozendict  # noqa
from gelidum.collections.frozenlist import frozenlist  # noqa
from gelidum.collections.frozenzet import frozenzet  # noqa

if NUMPY_INSTALLED:
    from gelidum.collections.frozenndarray import frozenndarray  # noqa



from typing import Any, Callable, Optional, Sequence, Generator, Union

from gelidum.exceptions import FrozenException
from gelidum.frozen import FrozenBase
from gelidum.typing import FrozenType, FrozenList

__all__ = ["frozenlist"]


_FrozenListParameterType = Optional[Union[Sequence, Generator]]


class frozenlist(tuple, FrozenBase):  # noqa
    def __raise_immutable_exception(self, *args, **kwargs):
        raise FrozenException("'frozenlist' object is immutable")

    def __new__(
        cls, seq: Optional[_FrozenListParameterType] = None, freeze_func: Optional[Callable[[Any], FrozenBase]] = None
    ) -> "frozenlist":
        if freeze_func is None:

            def freeze_func(item: Any) -> FrozenType:
                from gelidum.freeze import freeze

                return freeze(item, on_update="exception", on_freeze="copy")

        if seq:
            self = tuple.__new__(cls, (freeze_func(arg) for arg in seq))
        else:
            self = tuple.__new__(cls, [])
        return self

    def __init__(
        self, seq: Optional[_FrozenListParameterType] = None, freeze_func: Optional[Callable[[Any], FrozenBase]] = None
    ):
        pass

    @classmethod
    def _gelidum_on_update(cls, *args, **kwargs):
        raise FrozenException("'frozenlist' object is immutable")

    @classmethod
    def get_gelidum_hot_class_name(cls) -> str:
        return "tuple"

    @classmethod
    def get_gelidum_hot_class_module(cls) -> str:
        return "builtins.tuple"

    def __getitem__(self, key) -> Any:
        if type(key) is slice:
            return frozenlist(super().__getitem__(key))
        try:
            return super().__getitem__(key)
        except IndexError:
            raise IndexError("frozenlist index out of range")

    def __add__(self, other: FrozenList) -> FrozenList:
        joined_list = []
        for item in self:
            joined_list.append(item)
        for item in other:
            joined_list.append(item)
        return frozenlist(joined_list)

    def __mul__(self, times: int) -> FrozenList:
        as_list = []
        for item in self:
            as_list.append(item)
        return frozenlist(as_list * times)

    def append(self, item) -> None:
        self.__raise_immutable_exception()

    def extend(self, iterable):
        self.__raise_immutable_exception()

    def insert(self, i, x):
        self.__raise_immutable_exception()

    def remove(self, x):
        self.__raise_immutable_exception()

    def pop(self, items):
        self.__raise_immutable_exception()

    def clear(self):
        self.__raise_immutable_exception()

    def index(self, x, start=None, end=None) -> int:
        args = [x]
        if start:
            args.append(start)
        if end:
            args.append(end)
        try:
            return super().index(*args)
        except ValueError:
            raise ValueError(f"{x} is not in frozenlist")

    def sort(self, *, key=None, reverse=False):
        self.__raise_immutable_exception()

    def reverse(self):
        self.__raise_immutable_exception()

    def copy(self) -> "frozenlist":
        """
        frozendlist objects are only shallow-copied.
        """
        return self



import concurrent.futures
import unittest
from typing import Dict, List, Tuple, Any
from gelidum import freeze_params, freeze_final, FrozenException, Final


class TestDecorator(unittest.TestCase):
    def test_decorator_with_list_input_param(self):
        @freeze_params()
        def append_to_list(a_list: List, new_item: int):
            a_list.append(new_item)

        with self.assertRaises(FrozenException) as context:
            append_to_list([], 3)

        self.assertEqual("'frozenlist' object is immutable", str(context.exception))

    def test_decorator_with_dict_input_param(self):
        @freeze_params()
        def add_to_dict(a_dict: Dict, new_item: Tuple[str, Any]):
            a_dict[new_item[0]] = new_item[1]

        with self.assertRaises(TypeError) as context:
            add_to_dict({}, ("key", "value"))

        self.assertEqual("'frozendict' object is immutable", str(context.exception))

    def test_decorator_with_object_input_param_object_frozen(self):
        class Dummy(object):
            def __init__(self, attr: int):
                self.attr = attr

        @freeze_params()
        def assign_to_dummy(dummy_param: Dummy, value: int):
            dummy_param.attr = value

        dummy = Dummy(1)
        with self.assertRaises(TypeError) as context:
            assign_to_dummy(dummy, 999)

        dummy.attr = 100

        self.assertEqual("Can't assign attribute 'attr' on immutable instance", str(context.exception))
        self.assertEqual(100, dummy.attr)

    def test_decorator_only_some_params(self):
        class Dummy(object):
            def __init__(self, attr: int):
                self.attr = attr

        @freeze_params(params={"dummy_const"})
        def assign_to_dummies(a_dummy: Dummy, other_dummy: Dummy, the_dummy_const: Dummy):
            a_dummy.attr = the_dummy_const.attr
            other_dummy.attr = the_dummy_const.attr

        dummy1 = Dummy(1)
        dummy2 = Dummy(2)
        dummy_const = Dummy(99)
        assign_to_dummies(a_dummy=dummy1, other_dummy=dummy2, the_dummy_const=dummy_const)

        self.assertEqual(99, dummy1.attr)
        self.assertEqual(99, dummy2.attr)
        self.assertEqual(99, dummy_const.attr)

    def test_concurrent_futures(self):
        class Dummy(object):
            def __init__(self, attr: int):
                self.attr = attr

        @freeze_params()
        def job(dummy_input: Dummy):
            dummy_input.attr += 1

        thread_pool_executor = concurrent.futures.ThreadPoolExecutor(2)
        fut1 = thread_pool_executor.submit(job, Dummy(1))
        fut2 = thread_pool_executor.submit(job, Dummy(2))
        futures = [fut1, fut2]
        concurrent.futures.wait(futures)
        future_count = 0
        for future_i in futures:
            with self.assertRaises(FrozenException) as context:
                future_i.result()
            self.assertEqual("Can't assign attribute 'attr' on immutable instance", str(context.exception))
            future_count += 1
        self.assertEqual(2, future_count)

    def test_freeze_final_list_params(self):
        @freeze_final
        def join_lists_bad_implementation(one: Final[List], two: Final[List]):
            one.extend(two)
            return one  # pragma: no cover

        with self.assertRaises(FrozenException) as context_unnamed_arguments:
            join_lists_bad_implementation([], [])

        with self.assertRaises(FrozenException) as context_named_arguments:
            join_lists_bad_implementation(one=[], two=[])

        self.assertEqual("'frozenlist' object is immutable", str(context_unnamed_arguments.exception))
        self.assertEqual("'frozenlist' object is immutable", str(context_named_arguments.exception))

    def test_freeze_final_object_params(self):
        class Number(object):
            def __init__(self, value: int):
                self.value = value
                self.value_str = str(value)

        @freeze_final
        def product_bad_implementation(one: Final[Number], two: Final[Number], three: Number):
            three.value *= 99
            one.value *= two.value * three.value
            return one.value  # pragma: no cover

        with self.assertRaises(FrozenException) as context_unnamed_arguments:
            product_bad_implementation(Number(1), Number(2), Number(3))

        with self.assertRaises(FrozenException) as context_some_named_arguments:
            product_bad_implementation(Number(1), two=Number(2), three=Number(3))

        with self.assertRaises(FrozenException) as context_named_arguments:
            product_bad_implementation(one=Number(1), two=Number(2), three=Number(3))

        self.assertEqual(
            "Can't assign attribute 'value' on immutable instance", str(context_unnamed_arguments.exception)
        )
        self.assertEqual(
            "Can't assign attribute 'value' on immutable instance", str(context_some_named_arguments.exception)
        )
        self.assertEqual("Can't assign attribute 'value' on immutable instance", str(context_named_arguments.exception))
